import os
import h5py
from sklearn.model_selection import train_test_split
from keras.utils import to_categorical
import pickle
import multiprocessing as mp

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from astrorapid import helpers

# fix random seed for reproducibility
np.random.seed(42)


class PrepareArrays(object):
    def __init__(self, passbands=('g', 'r'), contextual_info=(0,)):
        self.passbands = passbands
        self.contextual_info = contextual_info
        self.nobs = 50
        self.npassbands = len(passbands)
        self.nfeatures = self.npassbands + len(self.contextual_info)
        self.timestep = 3.0
        self.mintime = -70
        self.maxtime = 80

    def make_cuts(self, data, i, deleterows, b, redshift=None, class_num=None, bcut=True, zcut=0.5, variables_cut=True):
        deleted = False
        time = data['r']['time'][0:self.nobs].dropna()
        if data.shape[0] < 4:
            print("Less than 4 epochs. nobs = {}".format(data.shape))
            deleterows.append(i)
            deleted = True
        elif len(time[time < 0]) < 3:
            print("Less than 3 points in the r band pre trigger", len(time[time < 0]))
            deleterows.append(i)
            deleted = True
        elif bcut and abs(b) < 15:
            print("In galactic plane. b = {}".format(b))
            deleterows.append(i)
            deleted = True
        elif zcut is not None and redshift is not None and redshift > self.zcut and redshift == 0:
            print("Redshift cut. z = {}".format(redshift))
            deleterows.append(i)
            deleted = True
        elif class_num is not None and variables_cut is True and class_num in [50, 70, 80, 81, 83, 84, 90, 91, 92, 93]:
            print("Not including variable models", class_num)
            deleterows.append(i)
            deleted = True

        return deleterows, deleted

    def get_min_max_time(self, data):
        # Get min and max times for tinterp
        mintimes = []
        maxtimes = []
        for j, pb in enumerate(self.passbands):
            if pb not in data:
                continue
            time = data[pb]['time'][0:self.nobs].dropna()
            mintimes.append(time.min())
            maxtimes.append(time.max())
        mintime = min(mintimes)
        maxtime = max(maxtimes) + self.timestep

        return mintime, maxtime

    def get_t_interp(self, data):
        mintime, maxtime = self.get_min_max_time(data)

        tinterp = np.arange(mintime, maxtime, step=self.timestep)
        len_t = len(tinterp)
        if len_t >= self.nobs:
            tinterp = tinterp[(tinterp >= self.mintime) & (tinterp <= self.maxtime)]
            len_t = len(tinterp)
            if len_t >= self.nobs:
                tinterp = tinterp[tinterp <= self.maxtime]
                len_t = len(tinterp)
                if len_t >= self.nobs:
                    tinterp = tinterp[len_t - self.nobs:]
                    len_t = len(tinterp)
        return tinterp, len_t

    def update_X(self, X, i, data, tinterp, len_t, objid, contextual_info, otherinfo):
        for j, pb in enumerate(self.passbands):
            if pb not in data:
                print("No", pb, "in objid:", objid)
                continue
            time = data[pb]['time'][0:self.nobs].dropna()
            try:
                flux = data[pb]['fluxRenorm'][0:self.nobs].dropna()
                fluxerr = data[pb]['fluxErrRenorm'][0:self.nobs].dropna()
            except KeyError:
                flux = data[pb][5][0:self.nobs].dropna()
                fluxerr = data[pb][6][0:self.nobs].dropna()
            photflag = data[pb]['photflag'][0:self.nobs].dropna()

            n = len(flux)  # Get vector length (could be less than nobs)

            if n > 1:
                if flux.values[-1] > flux.values[-2]:  # If last values are increasing, then set fill_values to zero
                    f = interp1d(time, flux, kind='linear', bounds_error=False, fill_value=0.)
                else:
                    f = interp1d(time, flux, kind='linear', bounds_error=False,
                                 fill_value='extrapolate')  # extrapolate until all passbands finished.

                fluxinterp = f(tinterp)
                fluxinterp = np.nan_to_num(fluxinterp)
                fluxinterp = fluxinterp.clip(min=0)
                X[i][j][0:len_t] = fluxinterp

        # Add contextual information
        for jj, c_idx in enumerate(contextual_info, 1):
            try:
                X[i][j + jj][0:len_t] = otherinfo[c_idx] * np.ones(len_t)
            except Exception as e:
                X[i][j + jj][0:len_t] = otherinfo[c_idx].values[0] * np.ones(len_t)

        return X


class PrepareInputArrays(PrepareArrays):
    def __init__(self, passbands=('g', 'r'), contextual_info=(0,), bcut=True, zcut=None):
        PrepareArrays.__init__(self, passbands, contextual_info)
        self.passbands = passbands
        self.contextual_info = contextual_info
        self.bcut = bcut
        self.zcut = zcut

    def prepare_input_arrays(self, lightcurves):
        nobjects = len(lightcurves)

        X = np.zeros(shape=(nobjects, self.nfeatures, self.nobs))
        timesX = np.zeros(shape=(nobjects, self.nobs))
        objids_list = []
        orig_lc = []
        deleterows = []

        for i, (objid, data) in enumerate(lightcurves.items()):
            print("Preparing light curve {} of {}".format(i, nobjects))

            otherinfo = data['otherinfo'].values.flatten()
            redshift, b, mwebv, trigger_mjd = otherinfo[0:4]

            # Make cuts
            deleterows, deleted = self.make_cuts(data, i, deleterows, b, redshift, class_num=None, bcut=self.bcut,
                                                 zcut=self.zcut)
            if deleted:
                continue

            tinterp, len_t = self.get_t_interp(data)
            timesX[i][0:len_t] = tinterp
            orig_lc.append(data)
            objids_list.append(objid)
            X = self.update_X(X, i, data, tinterp, len_t, objid, self.contextual_info, otherinfo)

        # Correct shape for keras is (N_objects, N_timesteps, N_passbands) (where N_timesteps is lookback time)
        X = X.swapaxes(2, 1)

        return X, orig_lc, timesX, objids_list


class PrepareTrainingSetArrays(PrepareArrays):
    def __init__(self, passbands=('g', 'r'), contextual_info=(0,), reread=False, aggregate_classes=False, bcut=True,
                 zcut=None, variablescut=False, nchunks=10000):
        PrepareArrays.__init__(self, passbands, contextual_info)
        self.passbands = passbands
        self.contextual_info = contextual_info
        self.reread = reread
        self.aggregate_classes = aggregate_classes
        self.bcut = bcut
        self.zcut = zcut
        self.variablescut = variablescut
        self.nchunks = nchunks
        self.agg_map = helpers.aggregate_sntypes()
        self.training_set_dir = 'training_set_files'
        if not os.path.exists(self.training_set_dir):
            os.makedirs(self.training_set_dir)

    def get_saved_light_curves_from_database(self, fpath_saved_lc):
        """ Get objids from hdf5 file

        Parameters
        ----------
        fpath_saved_lc : str
            File path of the light curves saved as a hdf5 file.

        Returns
        -------
        objids : list
            list of object IDs as strings.
        fpath_saved_lc : str
            Same as input argument

        """

        with h5py.File(fpath_saved_lc, 'r') as hdffile:
            objids = np.array(list(hdffile.keys()))
        np.random.shuffle(objids)

        return objids, fpath_saved_lc

    def prepare_training_set_arrays(self, fpath_saved_lc, otherchange=''):
        savepath = "X_{}ag{}_ci{}_fp{}_z{}_b{}_var{}.npy".format(otherchange, self.aggregate_classes,
                                                                 self.contextual_info, os.path.basename(self.fpath),
                                                                 self.zcut, self.bcut, self.variablescut)
        if self.reread is True or not os.path.isfile(os.path.join(self.training_set_dir, savepath)):
            objids = self.get_saved_light_curves_from_database(fpath_saved_lc)
            nobjects = len(objids)

            # Store data labels (y) and 'r' band data (X). Use memory mapping because input file is very large.
            labels = np.zeros(shape=nobjects, dtype=np.uint16)
            y = np.zeros(shape=(nobjects, self.nobs), dtype=np.uint16)
            X = np.memmap('X_lc_data.dat', dtype=np.float16, mode='w+', shape=(nobjects, self.nfeatures, self.nobs))
            X[:] = np.zeros(shape=(nobjects, self.nfeatures, self.nobs))
            timesX = np.zeros(shape=(nobjects, self.nobs))
            objids_list = []
            orig_lc = []

            # Chunk before multiprocesing
            multi_objids = np.array_split(objids, self.nchunks)

            # Store light curves into X (fluxes) and y (labels)
            pool = mp.Pool()
            results = pool.map_async(self.multi_read_obj, multi_objids)
            pool.close()
            pool.join()

            outputs = results.get()

            sum_deleterows = 0
            startidx = 0
            num_outputs = len(outputs)
            for i, output in enumerate(outputs):
                labels_part, y_part, X_part, timesX_part, objids_list_part, orig_lc_part, num_deleterows_part, num_objects_part = output
                endidx = startidx + num_objects_part
                labels[startidx:endidx] = labels_part
                y[startidx:endidx] = y_part
                X[startidx:endidx] = X_part
                timesX[startidx:endidx] = timesX_part
                objids_list.extend(objids_list_part)
                orig_lc.extend(orig_lc_part)
                startidx += num_objects_part
                sum_deleterows += num_deleterows_part
                print('combining results...', i, num_outputs)

            deleterows = np.array(np.arange(nobjects - sum_deleterows, nobjects))
            X = np.delete(X, deleterows, axis=0)
            y = np.delete(y, deleterows, axis=0)
            labels = np.delete(labels, deleterows, axis=0)
            timesX = np.delete(timesX, deleterows, axis=0)

            np.save(
                "X_{}ag{}_ci{}_fp{}_z{}_b{}_var{}.npy".format(otherchange, self.aggregate_classes, self.contextual_info,
                                                              os.path.basename(fpath_saved_lc), self.zcut, self.bcut,
                                                              self.variablescut), X)
            np.save(
                "y_{}ag{}_ci{}_fp{}_z{}_b{}_var{}.npy".format(otherchange, self.aggregate_classes, self.contextual_info,
                                                              os.path.basename(fpath_saved_lc), self.zcut, self.bcut,
                                                              self.variablescut), y)
            np.save("labels_{}ag{}_ci{}_fp{}_z{}_b{}_var{}.npy".format(otherchange, self.aggregate_classes,
                                                                       self.contextual_info,
                                                                       os.path.basename(fpath_saved_lc), self.zcut,
                                                                       self.bcut, self.variablescut), labels)
            np.save("tinterp_{}ag{}_ci{}_fp{}_z{}_b{}_var{}.npy".format(otherchange, self.aggregate_classes,
                                                                        self.contextual_info,
                                                                        os.path.basename(fpath_saved_lc), self.zcut,
                                                                        self.bcut, self.variablescut), timesX)
            np.save("objids_{}ag{}_ci{}_fp{}_z{}_b{}_var{}.npy".format(otherchange, self.aggregate_classes,
                                                                       self.contextual_info,
                                                                       os.path.basename(fpath_saved_lc), self.zcut,
                                                                       self.bcut, self.variablescut), objids_list)
            with open("origlc_{}ag{}_ci{}_fp{}_z{}_b{}_var{}.npy".format(otherchange, self.aggregate_classes,
                                                                         self.contextual_info,
                                                                         os.path.basename(fpath_saved_lc), self.zcut,
                                                                         self.bcut, self.variablescut), 'wb') as f:
                pickle.dump(orig_lc, f)

        else:
            self.X = np.load(
                "X_{}ag{}_ci{}_fp{}_z{}_b{}_var{}.npy".format(otherchange, self.aggregate_classes, self.contextual_info,
                                                              os.path.basename(fpath_saved_lc), self.zcut, self.bcut,
                                                              self.variablescut), mmap_mode='r')
            self.y = np.load(
                "y_{}ag{}_ci{}_fp{}_z{}_b{}_var{}.npy".format(otherchange, self.aggregate_classes, self.contextual_info,
                                                              os.path.basename(fpath_saved_lc), self.zcut, self.bcut,
                                                              self.variablescut))
            self.labels = np.load(
                "labels_{}ag{}_ci{}_fp{}_z{}_b{}_var{}.npy".format(otherchange, self.aggregate_classes,
                                                                   self.contextual_info,
                                                                   os.path.basename(fpath_saved_lc), self.zcut,
                                                                   self.bcut, self.variablescut))
            self.timesX = np.load(
                "tinterp_{}ag{}_ci{}_fp{}_z{}_b{}_var{}.npy".format(otherchange, self.aggregate_classes,
                                                                    self.contextual_info,
                                                                    os.path.basename(fpath_saved_lc), self.zcut,
                                                                    self.bcut, self.variablescut))
            self.objids_list = np.load(
                "objids_{}ag{}_ci{}_fp{}_z{}_b{}_var{}.npy".format(otherchange, self.aggregate_classes,
                                                                   self.contextual_info,
                                                                   os.path.basename(fpath_saved_lc), self.zcut,
                                                                   self.bcut, self.variablescut))
            with open("origlc_{}ag{}_ci{}_fp{}_z{}_b{}_var{}.npy".format(otherchange, self.aggregate_classes,
                                                                         self.contextual_info,
                                                                         os.path.basename(fpath_saved_lc), self.zcut,
                                                                         self.bcut, self.variablescut), 'rb') as f:
                orig_lc = pickle.load(f)

        classes = sorted(list(set(self.labels)))
        sntypes_map = helpers.get_sntypes()
        class_names = [sntypes_map[class_num] for class_num in classes]

        # Count nobjects per class
        for c in classes:
            nobs = len(self.X[self.labels == c])
            print(c, nobs)

        # Use class numbers 1,2,3... instead of 1, 3, 13 etc.
        y_indexes = np.copy(self.y)
        for i, c in enumerate(classes):
            y_indexes[self.y == c] = i + 1
        self.y = y_indexes

        self.y = to_categorical(self.y)

        # Correct shape for keras is (N_objects, N_timesteps, N_passbands) (where N_timesteps is lookback time)
        self.X = self.X.swapaxes(2, 1)

        X_train, X_test, y_train, y_test, labels_train, labels_test, timesX_train, timesX_test, orig_lc_train, orig_lc_test, objids_train, objids_test = train_test_split(
            self.X, self.y, self.labels, self.timesX, self.orig_lc, self.objids_list, train_size=0.60,
            shuffle=False, random_state=42)

        counts = np.unique(labels_train, return_counts=True)[-1]
        class_weights = max(counts) / counts
        class_weights = dict(zip(range(len(counts)), class_weights))
        print("Class weights:", class_weights)

        # Sample weights
        l_train_indexes = np.copy(labels_train)
        for i, c in enumerate(classes):
            l_train_indexes[l_train_indexes == c] = i
        sample_weights = np.zeros(len(l_train_indexes))
        for key, val in class_weights.items():
            sample_weights[l_train_indexes == key] = val

        return X_train, X_test, y_train, y_test, labels_train, labels_test, class_names, class_weights, \
               sample_weights, timesX_train, timesX_test, orig_lc_train, orig_lc_test, objids_train, objids_test

    def multi_read_obj(self, objids):
        nobjects = len(objids)

        labels = np.zeros(shape=nobjects, dtype=np.uint16)
        y = np.zeros(shape=(nobjects, self.nobs), dtype=np.uint16)
        X = np.zeros(shape=(nobjects, self.nfeatures, self.nobs))
        timesX = np.zeros(shape=(nobjects, self.nobs))
        objids_list = []
        orig_lc = []
        deleterows = []

        for i, objid in enumerate(objids):
            print("Preparing {} light curve {} of {}".format(objid, i, self.nobjects))

            # Get aggregate model
            field, model, base, snid = objid.astype(str).split('_')
            if self.aggregate_classes:
                model = self.agg_map[int(model)]
            class_num = int(model)

            # Get data for each object
            try:
                data = pd.read_hdf(self.fpath, key=objid)
            except AttributeError:
                print("ignoring: cannot read", objid)
                deleterows.append(i)
                continue

            otherinfo = data['otherinfo'].values.flatten()
            redshift, b, mwebv, trigger_mjd, t0, peakmjd = otherinfo[0:6]

            # Make cuts
            deleterows, deleted = self.make_cuts(data, i, deleterows, b, redshift, class_num=model, bcut=self.bcut,
                                                 zcut=self.zcut, variables_cut=self.variablescut)
            if deleted:
                continue

            tinterp, len_t = self.get_t_interp(data)
            timesX[i][0:len_t] = tinterp
            orig_lc.append(data)
            objids_list.append(objid)
            X = self.update_X(X, i, data, tinterp, len_t, objid, self.contextual_info, otherinfo)

            activeindexes = (tinterp > t0)
            labels[i] = int(model)
            y[i][0:len_t] = int(model) * activeindexes

        deleterows = np.array(deleterows)
        X = np.delete(X, deleterows, axis=0)
        y = np.delete(y, deleterows, axis=0)
        labels = np.delete(labels, deleterows, axis=0)
        timesX = np.delete(timesX, deleterows, axis=0)
        count_deleterows = len(deleterows)
        num_objects = X.shape[0]

        return labels, y, X, timesX, objids_list, orig_lc, count_deleterows, num_objects



from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.cosmology import WMAP9 as cosmo
import numpy as np
import pandas as pd

from astrorapid import helpers
from astrorapid.ANTARES_object.LAobject import LAobject
from astrorapid.ANTARES_object.features import model_early_lightcurve


class InputLightCurve(object):
    def __init__(self, mjd, flux, fluxerr, passband, zeropoint, photflag, ra, dec, objid, redshift=None, mwebv=None,
                 known_redshift=True, training_set_parameters=None):
        """

        Parameters
        ----------
        mjd : array
            List of floating point mjd times for entire light curve in all passbands.
        flux : array
            List of floating point fluxes for entire light curve in all passbands.
        fluxerr : array
            List of floating point flux errors times for entire light curve in all passbands.
        passband : array
            List of passbands as strings for each point in the array.
        zeropoint : array
            List of floating point zeropoints for entire light curve in all passbands.
        photflag : array
            List of integer flags indicating whether the observation was a detection or non-detection for each
            point in the light curve.
        ra : float
            Right Ascension in degrees.
        dec : float
            Declination in degrees.
        objid : str
            String to identify the input object.
        redshift : float
            Optional parameter. Cosmological redshift.
        mwebv : float
            Optional parameter. Milky Way E(B-V) extinction.
        known_redshift : bool
            Whether to use redshift in processing the light curves and making the arrays.
        training_set_parameters : dict
            Optional parameter. If this is not None, then determine the explosion time, t0, for full the training set.
            The dictionary must have the following keys: {class_number, peakmjd}
        """

        self.mjd = np.array(mjd)
        self.flux = np.array(flux)
        self.fluxerr = np.array(fluxerr)
        self.passband = np.array(passband)
        self.zeropoint = np.array(zeropoint)
        self.photflag = np.array(photflag)
        self.ra = ra
        self.dec = dec
        self.objid = objid
        self.redshift = redshift
        self.mwebv = mwebv
        self.known_redshift =known_redshift
        self.training_set_parameters = training_set_parameters
        if training_set_parameters is not None:
            self.class_number = training_set_parameters['class_number']
            self.peakmjd = training_set_parameters['peakmjd']

        self.b = self.get_galactic_latitude()
        self.trigger_mjd, self.t = self.get_trigger_time()

    def get_galactic_latitude(self):
        c_icrs = SkyCoord(ra=self.ra * u.degree, dec=self.dec * u.degree, frame='icrs')
        b = c_icrs.galactic.b.value

        return b

    def get_trigger_time(self):
        trigger_mjd = float(self.mjd[self.photflag == 6144][0])
        t = self.mjd - trigger_mjd

        return trigger_mjd, t

    def correct_time_dilation(self, t):
        t = t / (1 + self.redshift)

        return t

    def correct_for_distance(self, flux, fluxerr):
        dlmu = cosmo.distmod(self.redshift).value
        flux, fluxerr = helpers.calc_luminosity(flux, fluxerr, dlmu)

        return flux, fluxerr

    def compute_t0(self, outlc):
        """ Calculate the explosion time for the trianing set if certain conditions are met. """
        calc_params = True
        inrange_mask = self.t < self.peakmjd
        if int(self.class_number) in [70, 80, 82, 83]:  # No t0 if model types (AGN, RRlyrae, Eclipsing Binaries)
            calc_params = False
        elif self.peakmjd - self.trigger_mjd < 0:  # No t0 if trigger to peak time is small
            calc_params = False
        elif len(self.t[inrange_mask]) < 3:  # No t0 if not At least 3 points before peak
            calc_params = False
        elif len(self.t[self.t < 0]) < 3:  # No t0 if there are less than three points before trigger. This suggests that it triggered on one of the first point of a passband
            calc_params = False

        if calc_params:
            earlytime = self.peakmjd - self.trigger_mjd
            fit_func, parameters = model_early_lightcurve.fit_early_lightcurve(outlc, earlytime)
        else:
            parameters = {pb: [-99, -99, -99] for pb in self.passband}

        t0 = parameters[next(iter(parameters))][2]  # Get t0 from any of the passbands

        return t0

    def preprocess_light_curve(self):
        """ Preprocess light curve. """

        # Account for distance and time dilation if redshift is known
        if self.known_redshift and self.redshift is not None:
            self.t = self.correct_time_dilation(self.t)
            self.flux, self.fluxerr = self.correct_for_distance(self.flux, self.fluxerr)

        obsid = np.arange(len(self.t))

        laobject = LAobject(locusId=self.objid, objectId=self.objid, time=self.t, flux=self.flux, fluxErr=self.fluxerr,
                            obsId=obsid, passband=self.passband, zeropoint=self.zeropoint, per=False, mag=False,
                            photflag=self.photflag, z=self.redshift)

        outlc = laobject.get_lc(recompute=True)

        otherinfo = [self.redshift, self.b, self.mwebv, self.trigger_mjd]

        if self.training_set_parameters is not None:
            t0 = self.compute_t0(outlc)
            otherinfo += [t0, self.peakmjd]

        savepd = {pb: pd.DataFrame(lcinfo).loc[[0, 5, 6, 7]].rename({0: 'time', 5: 'fluxRenorm', 6: 'fluxErrRenorm', 7: 'photflag'}).T for pb, lcinfo in outlc.items()}  # Convert to dataframe rows: time, fluxNorm, fluxNormErr, photFlag; columns: ugrizY
        savepd['otherinfo'] = pd.DataFrame(otherinfo)
        savepd = pd.DataFrame(
            {(outerKey, innerKey): values for outerKey, innerDict in savepd.items() for innerKey, values in
             innerDict.items()})  # Use multilevel indexing

        return savepd


def read_multiple_light_curves(light_curve_list, known_redshift=True, training_set_parameters=None):
    """
    light_curve_list is a list of tuples with each tuple having entries:
    mjd, flux, fluxerr, passband, zeropoint, photflag, ra, dec, objid, redshift, mwebv

    Returns the processed light curves
    """

    processed_light_curves = {}
    for light_curve in light_curve_list:
        mjd, flux, fluxerr, passband, zeropoint, photflag, ra, dec, objid, redshift, mwebv = light_curve
        inputlightcurve = InputLightCurve(*light_curve, known_redshift, training_set_parameters)
        processed_light_curves[objid] = inputlightcurve.preprocess_light_curve()

    return processed_light_curves









import numpy as np
from scipy.optimize import curve_fit, minimize
import matplotlib.pyplot as plt
import emcee
from scipy.stats import chisquare
import pylab
from collections import OrderedDict
from matplotlib import rc
import matplotlib as mpl
# mpl.use('GTK3AGG')
rc('text', usetex=False)
# rc('axes', unicode_minus=True)
# rc('font', **{'family':'serif','serif':['Times New Roman']})
from . import model_early_lightcurve

class EarlyMixin(object):
    """
    Methods to derive early classification features for LAobjects
    """

    def get_early_rise_rate(self, recompute=False, plot=False, passbands=('u', 'g', 'r', 'i', 'z', 'Y')):
        """
        Compute the early rise rate (slope) of the light curve.
        """

        earlyriserate = getattr(self, 'earlyriserate', None)
        a_fit = getattr(self, 'a_fit', None)
        c_fit = getattr(self, 'c_fit', None)
        if earlyriserate is not None:
            if not recompute:
                return earlyriserate, a_fit, c_fit

        earlyriserate = {}
        a_fit = {}
        c_fit = {}
        return_vals = {}
        outlc = self.get_lc(recompute=recompute)

        fit_func = getattr(self, 'early_fit_func', None)
        parameters = getattr(self, 'early_parameters', None)
        if fit_func is None or parameters is None:
            fit_func, parameters = model_early_lightcurve.fit_early_lightcurve(outlc)
            self.early_fit_func, self.early_parameters = fit_func, parameters

        if plot:
            plt.figure(figsize=(12, 10))
            ymin, ymax = -10, 10

        col = {'u': 'b', 'g': 'g', 'r': 'r', 'i': 'm', 'z': 'k', 'Y': 'y'}
        for i, pb in enumerate(passbands):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc

            tFluxUnred = tFluxUnred - np.median(tFluxUnred[ttime < 0])
            # mask = ttime > -30
            # tFluxUnred = tFluxUnred[mask]
            # ttime = ttime[mask]
            # tFluxErrUnred = tFluxErrUnred[mask]

            if len(ttime) <= 1 or not np.any(ttime < 0):
                return_vals[pb] = (-99, -99, -99)
                if plot:
                    plt.errorbar(ttime, tFluxUnred, yerr=tFluxErrUnred, fmt='.', color=col[pb], label=pb)
                continue

            fit_flux = fit_func(np.arange(min(ttime), max(ttime), 0.2), *parameters[pb])

            a, c, t0 = parameters[pb]

            t1 = 1 + t0
            t2 = 10 + t0
            f1 = fit_func(t1, *parameters[pb])
            f2 = fit_func(t2, *parameters[pb])
            tearlyriserate = (f2 - f1) / (t2 - t1)

            earlyriserate[pb] = tearlyriserate
            a_fit[pb] = parameters[pb][0]
            c_fit[pb] = parameters[pb][1]
            return_vals[pb] = (tearlyriserate, parameters[pb][0], parameters[pb][1])

            if plot:
                label = "{:>2}: a ={:6.2f}, c ={:6.2f}, slope ={:6.2f}".format(pb, a, c, tearlyriserate)
                plt.errorbar(ttime, tFluxUnred, yerr=tFluxErrUnred, fmt='.', color=col[pb], label=label)
                plt.plot(np.arange(min(ttime), max(ttime), 0.2), fit_flux, color=col[pb])
                ymin = min(ymin, (min(tFluxUnred[ttime >= -40]) - max(tFluxErrUnred[ttime >= -40])))
                ymax = max(ymax, (max(tFluxUnred[ttime >= -40]) + max(tFluxErrUnred[ttime >= -40])))

        if plot:
            plt.title(self.objectId)
            plt.xlabel("Days since trigger (rest frame)", fontsize=13)
            plt.ylabel("Relative Flux (distance corrected)", fontsize=13)
            plt.xlim(-40, 15)
            plt.ylim(ymin, ymax)
            plt.autoscale(enable=True, axis='y', tight=True)
            plt.tick_params(axis='both', labelsize=12)

            plt.errorbar([0], [0], yerr=[0], alpha=0.0, label="$t_0$ ={:7.2f}".format(t0))
            plt.errorbar([0], [0], yerr=[0], alpha=0.0, label="$z = {:.3f}$".format(self.z))
            # plt.autoscale(enable=True, axis='y', tight=True)
            plt.legend(frameon=False, loc='upper left')
            plt.savefig(self.objectId + "_ylims_fit.pdf")
            plt.show()

        self.earlyriserate = earlyriserate
        self.a_fit = a_fit
        self.c_fit = c_fit
        return return_vals

    def get_color_at_n_days(self, n, recompute=True, passbands=('u', 'g', 'r', 'i', 'z', 'Y')):
        """
        Compute the colors at n days and the linear slope of the color
        passbands = ('u', 'g', 'r', 'i', 'z', 'Y')  # order of filters matters as it must be 'u-g' rather than 'g-u'
        """
        color = getattr(self, 'color', None)
        if color is not None:
            if not recompute:
                return color

        color = {}
        color_slope = {}
        outlc = self.get_lc(recompute=recompute)


        fit_func = getattr(self, 'early_fit_func', None)
        parameters = getattr(self, 'early_parameters', None)
        if fit_func is None or parameters is None:
            fit_func, parameters = model_early_lightcurve.fit_early_lightcurve(outlc)
            self.early_fit_func, self.early_parameters = fit_func, parameters

        ignorepb = []
        tflux_ndays = {}
        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc

            if len(ttime) <= 1 or not np.any(ttime < 0):
                tflux_ndays[pb] = 0
                continue

            # Check if there is data after t0 trigger for this passband
            a, c, t0 = parameters[pb]
            if max(ttime) < t0:
                print('No data for', pb, ttime)
                ignorepb.append(pb)
            else:
                print("time", pb, ttime)

            n = t0 + n

            tflux_ndays[pb] = fit_func(n, *parameters[pb])

        # plt.figure()

        for i, pb1 in enumerate(passbands):
            for j, pb2 in enumerate(passbands):
                if i < j:
                    c = pb1 + '-' + pb2
                    if pb1 not in parameters.keys() or pb2 not in parameters.keys() or tflux_ndays[pb2] == 0 or (pb1 in ignorepb or pb2 in ignorepb):
                        color[c] = -99.
                        color_slope[c] = -99.
                        # print("Not plotting", c)
                        continue
                    color[c] = -2.5*np.log10(tflux_ndays[pb1] / tflux_ndays[pb2])
                    color_slope[c] = color[c] / n
                    # fit_t = np.arange(-30, 15, 0.2)
                    # plotcolor = -2.5*np.log10(fit_func(fit_t, *parameters[pb1])/fit_func(fit_t, *parameters[pb2]))
                    # plt.plot(fit_t[fit_t >= t0], plotcolor[fit_t >= t0], label=c)

        # # plt.title(self.objectId)
        # plt.xlabel("Days since trigger (rest frame)")
        # plt.ylabel("Color")
        # # plt.xlim(-40, 15)
        # plt.legend(frameon=False)
        # plt.show()
        self.color = color
        self.color_slope = color_slope
        return color, color_slope


def lnlike(params, times, fluxes, fluxerrs):
        # print(params)
        t0 = params[0]
        pars = params[1:]

        chi2 = 0
        for i, pb in enumerate(times):
            a, c = pars[i * 2:i * 2 + 2]

            model = np.heaviside((times[pb] - t0), 1) * (a * (times[pb] - t0) ** 2) + c
            chi2 += sum((fluxes[pb] - model) ** 2 / fluxerrs[pb] ** 2)

            # print(pb, a, c)

        # print('chi2', chi2, params)
        return -chi2


def lnprior(params):
    # print('params', params)
    t0 = params[0]
    pars = params[1:]

    if -35 < t0 < 0:
        return 0.0
    return -np.inf


def lnprob(params, t, flux, fluxerr):
    lp = lnprior(params)
    if not np.isfinite(lp):
        return -np.inf
    return lp + lnlike(params, t, flux, fluxerr)


def emcee_fit_all_pb_lightcurves(times, fluxes, fluxerrs, ndim, x0=None, bounds=None):
    nwalkers = 200
    nsteps = 700
    burn = 50
    pos = np.array([x0 + 3*np.random.randn(ndim) for i in range(nwalkers)])
    # print(pos[:,0])

    # print("DATA IS TO FOLLOW")
    # print("times")
    # print(times)
    # print("fluxes")
    # print(fluxes)
    # print("fluxerrs")
    # print(fluxerrs)

    print("running mcmc...")

    # Ensure intial params within parameter bounds
    params = OrderedDict()
    params['t0'] = {'bounds': bounds[0], 'value': x0[0], 'scale': 1}
    i = 1
    for pb in times.keys():
        for name in ['a', 'c']:
            params[pb + ': ' + name] = {'bounds': bounds[i], 'value': x0[i], 'scale': 3}
            i += 1
    for i, name in enumerate(params.keys()):
        # import pdb; pdb.set_trace()
        lb, ub = params[name]['bounds']
        p0     = params[name]['value']
        std    = params[name]['scale']
        # take a 5 sigma range
        lr, ur = (p0-5.*std, p0+5.*std)
        ll = max(lb, lr, 0.)
        ul = min(ub, ur)
        ind = np.where((pos[:,i] <= ll) | (pos[:,i] >= ul))
        nreplace = len(pos[:,i][ind])
        pos[:,i][ind] = np.random.rand(nreplace)*(ul - ll) + ll
    # print('new', np.array(pos)[:,0])
    # print(pos)

    sampler = emcee.EnsembleSampler(nwalkers, ndim, lnprob, args=(times, fluxes, fluxerrs))

    # FInd parameters of lowest chi2
    pos, prob, state = sampler.run_mcmc(pos, 1)
    opos, oprob, orstate = [], [], []
    for pos, prob, rstate in sampler.sample(pos, prob, state, iterations=nsteps):
        opos.append(pos.copy())
        oprob.append(prob.copy())
    pos = np.array(opos)
    prob = np.array(oprob)
    nstep, nwalk = np.unravel_index(prob.argmax(), prob.shape)
    bestpars = pos[nstep, nwalk]
    posterior=prob

    print("best", bestpars)
    # import pylab
    # for j in range(nwalkers):
    #     pylab.plot(posterior[:, j])
    # for i in range(len(params)):
    #     pylab.figure()
    #     pylab.title(list(params.keys())[i])
    #     for j in range(nwalkers):
    #         pylab.plot(pos[:, j, i])

    samples = sampler.chain[:, burn:, :].reshape((-1, ndim))
    # samples[:, 2] = np.exp(samples[:, 2])
    # bestpars2 = list(map(lambda v: (v[0]), zip(*np.percentile(samples, [50], axis=0))))
    # print('b2', bestpars2)

    print(bestpars)
    t0 = bestpars[0]
    bestpars = bestpars[1:]

    # for i, pb in enumerate(times):
    #     if len(times[pb][times[pb] > t0]) == 0:
    #         bestpars[i * 2] = 0  # set `a' to zero if no data points above t0

    best = {pb: np.append(bestpars[i * 2:i * 2 + 2], t0) for i, pb in enumerate(times)}

    # to_delete = []
    # for i, name in enumerate(params):
    #     if not np.any(samples[:, i]):  # all zeros for parameter then delete column
    #         to_delete.append((i, name))
    # for i, name in to_delete:
    #     samples = np.delete(samples, i, axis=1)
    #     print(name)
    #     del params[name]
    #
    # np.save('samples', samples[::100])

    return best



# -*- coding: UTF-8 -*-
"""
Functions for Gaussian Process Regression with  LAobjects

"""

from __future__ import absolute_import
from __future__ import unicode_literals
import numpy as np
import george
from george import kernels
import scipy.optimize as op


class GPMixin(object):
    """
    Methods to smooth lightcurves with a Gaussian Process
    """
    def gaussian_process_smooth(self, per=None, minobs=10, phase_offset=None, recompute=False, scalemin=None, scalemax=None):
        """
        per = cheaty hackish thing to get a gaussian process with some
        continuity at the end points

        minobs = mininum number of observations in each filter before we fit
        """
        outgp = getattr(self, 'outgp', None)

        if outgp is not None:
            if not recompute:
                return outgp
        else:
            outgp = {}

        # note that we explicitly ask for non-smoothed and non-gpr lc here, since we're trying to compute the gpr
        outlc = self.get_lc(recompute=False, per=per, smoothed=False, gpr=False, phase_offset=phase_offset)

        for i, pb in enumerate(outlc):
            thislc = outlc.get(pb)
            thisphase, thisFlux, thisFluxErr = thislc

            nobs = len(thisphase)

            # if we do not have enough observations in this passband, skip it
            if nobs < minobs:
                continue

            # TODO : REVISIT KERNEL CHOICE
            if per == 1:
                # periodic
                kernel = kernels.ConstantKernel(1.)*kernels.ExpSine2Kernel(1.0, 0.0)
            elif per == 2:
                # quasiperiodic
                kernel = kernels.ConstantKernel(1.)*kernels.ExpSquaredKernel(100.)*kernels.ExpSine2Kernel(1.0, 0.0)
            else:
                # non-periodic
                kernel = kernels.ConstantKernel(1.)*kernels.ExpSquaredKernel(1.)

            gp = george.GP(kernel, mean=thisFlux.mean(), fit_mean=True, fit_white_noise=True, white_noise=np.log(thisFluxErr.mean()**2.))

            # define the objective function
            def nll(p):
                gp.set_parameter_vector(p)
                ll = gp.lnlikelihood(thisFlux, quiet=True)
                return -ll if np.isfinite(ll) else 1e25

            # define the gradient of the objective function.
            def grad_nll(p):
                gp.set_parameter_vector(p)
                return -gp.grad_lnlikelihood(thisFlux, quiet=True)

            # pre-compute the kernel
            gp.compute(thisphase, thisFluxErr)
            p0 = gp.get_parameter_vector()

            max_white_noise = np.log((3.*np.median(thisFluxErr))**2.)
            min_white_noise = np.log((0.3*np.median(thisFluxErr))**2)

            # coarse optimization with scipy.optimize
            # TODO : almost anything is better than scipy.optimize
            if per ==1:
                # mean, white_noise, amplitude, gamma, FIXED_period
                results = op.minimize(nll, p0, jac=grad_nll, bounds=[(None, None), (min_white_noise, max_white_noise),\
                        (None, None), (None, None), (0.,0.)])
            elif per==2:
                # mean white_noise, amplitude, variation_timescale, gamma, FIXED_period
                results = op.minimize(nll, p0, jac=grad_nll, bounds=[(None, None), (min_white_noise, max_white_noise),\
                        (None, None), (scalemin, scalemax), (None, None), (0.,0.)])
            else:
                # mean white_noise, amplitude, variation_timescale
                results = op.minimize(nll, p0, jac=grad_nll, bounds=[(None, None), (min_white_noise, max_white_noise),\
                        (None, None), (scalemin,scalemax)])

            gp.set_parameter_vector(results.x)
            # george is a little different than sklearn in that the prediction stage needs the input data
            outgp[pb] = (gp, thisphase, thisFlux, thisFluxErr)

        self.outgp = outgp
        return outgp



import os
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from scipy.interpolate import interp1d
from keras.utils import to_categorical
from sklearn.metrics import confusion_matrix
import imageio
import matplotlib

from astrorapid.classifier_metrics import plot_confusion_matrix, compute_multiclass_roc_auc, compute_precision_recall, plasticc_log_loss


def plot_metrics(class_names, model, X_test, y_test, fig_dir, timesX_test=None, orig_lc_test=None, objids_test=None, passbands=('g', 'r')):
    scores = model.evaluate(X_test, y_test, verbose=0)
    print("Accuracy: %.2f%%" % (scores[1] * 100))

    y_pred = model.predict(X_test)
    y_test_indexes = np.argmax(y_test, axis=-1)
    y_pred_indexes = np.argmax(y_pred, axis=-1)

    accuracy = len(np.where(y_pred_indexes == y_test_indexes)[0])
    print("Accuracy is: {}/{} = {}".format(accuracy, len(y_test_indexes.flatten()), accuracy / len(y_test_indexes.flatten())))

    class_names = ["Pre-explosion"] + class_names

    timesX_test[timesX_test == 0] = -200

    for cname in class_names:
        dirname = os.path.join(fig_dir + '/lc_pred', cname)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

    plt.rcParams['text.usetex'] = True
    plt.rcParams['font.serif'] = ['Computer Modern Roman'] + plt.rcParams['font.serif']

    # Plot accuracy vs time per class
    font = {'family': 'normal',
            'size': 36}
    matplotlib.rc('font', **font)
    print("Plotting Accuracy vs time per class...")
    fig = plt.figure("accuracy_vs_time_perclass", figsize=(13, 12))
    # fig = plt.figure(figsize=(13, 12))
    for classnum, classname in enumerate(class_names):
        correct_predictions_inclass = (y_test_indexes == classnum) & (y_pred_indexes == y_test_indexes)
        time_bins = np.arange(-110, 110, 3.)
        times_binned_indexes = np.digitize(timesX_test, bins=time_bins, right=True)
        time_list_indexes_inclass, count_correct_vs_binned_time_inclass = np.unique(
            times_binned_indexes * correct_predictions_inclass, return_counts=True)
        time_list_indexes2_inclass, count_objects_vs_binned_time_inclass = np.unique(
            times_binned_indexes * (y_test_indexes == classnum), return_counts=True)
        start_time_index = int(np.where(time_list_indexes2_inclass == time_list_indexes_inclass[1])[0])
        end_time_index = int(np.where(time_list_indexes2_inclass == time_list_indexes_inclass[-1])[0]) + 1

        try:
            accuracy_vs_time_inclass = count_correct_vs_binned_time_inclass[1:] / count_objects_vs_binned_time_inclass[
                                                                                  start_time_index:end_time_index]
        except Exception as e:
            print(e)
            continue

        try:
            assert time_list_indexes_inclass[1:] == time_list_indexes2_inclass[start_time_index:end_time_index]
        except:
            pass

        if classname == 'Non-detection' or classname == 'Pre-explosion':
            pass  # plt.plot(time_bins[time_list_indexes_inclass[1:]], accuracy_vs_time_inclass, '-', label=classname, alpha=0.6, color=COLORS[classnum])
        else:
            plt.plot(time_bins[time_list_indexes_inclass[1:]], accuracy_vs_time_inclass, '-', label=classname,
                     color=COLORS[classnum], lw=3)
    plt.xlim(left=-35, right=70)
    plt.xlabel("Days since trigger (rest frame)")
    plt.ylabel("Classification accuracy")
    plt.legend(frameon=True, fontsize=25, ncol=2, loc=0)  # , bbox_to_anchor=(0.05, -0.1))
    plt.tight_layout()
    plt.savefig(os.path.join(fig_dir, "accuracy_vs_time_perclass.pdf"), bbox_inches='tight')
    plt.close()

    font = {'family': 'normal',
            'size': 33}

    matplotlib.rc('font', **font)

    # Plot confusion matrix at different days past trigger
    print("Plotting Confusion Matrices...")
    time_bins = np.arange(-110, 110, 1.)
    nobjects = len(timesX_test)
    ntimesteps = len(time_bins)
    nclasses = y_test.shape[-1]
    y_test_indexes_days_past_trigger = np.zeros((nobjects, ntimesteps))
    y_pred_indexes_days_past_trigger = np.zeros((nobjects, ntimesteps))
    y_pred_days_past_trigger = np.zeros((nobjects, ntimesteps, nclasses))
    for objidx in range(nobjects):
        print(objidx, nobjects, 'For conf matrix')
        f = interp1d(timesX_test[objidx], y_test_indexes[objidx], kind='nearest', bounds_error=False,
                     fill_value='extrapolate')
        y_test_indexes_days_past_trigger[objidx][:] = f(time_bins)
        f = interp1d(timesX_test[objidx], y_pred_indexes[objidx], kind='nearest', bounds_error=False,
                     fill_value='extrapolate')
        y_pred_indexes_days_past_trigger[objidx][:] = f(time_bins)
        for classidx in range(nclasses):
            classprob = y_pred[objidx][:, classidx]
            mintimeidx = 0
            maxtimeidx = np.argmax(timesX_test[objidx])
            f = interp1d(timesX_test[objidx], classprob, kind='linear', bounds_error=False,
                         fill_value=(classprob[mintimeidx], classprob[maxtimeidx]))
            y_pred_days_past_trigger[objidx][:, classidx][:] = f(time_bins)

    images_cf, images_roc, images_pr = [], [], []
    roc_auc, pr_auc = {}, {}
    wlogloss = {}
    for i, days_since_trigger in enumerate(list(np.arange(-25, 25)) + list(np.arange(25, 70, step=5))):  # [('early2days', 2), ('late40days', 40)]: # [-25, -20, -15, -10, -5, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70]:
        print("Plotting CF matrix", i, days_since_trigger, "days")
        index = np.where(time_bins == days_since_trigger)[0][0]
        y_test_on_day_i = y_test_indexes_days_past_trigger[:, index]
        y_pred_on_day_i = y_pred_indexes_days_past_trigger[:, index]
        y_pred_prob_on_day_i = y_pred_days_past_trigger[:, index]

        name = 'since_trigger_{}'.format(days_since_trigger)
        title = '{} days since trigger'.format(days_since_trigger)
        cnf_matrix = confusion_matrix(y_test_on_day_i, y_pred_on_day_i)
        print(name, cnf_matrix)
        figname_cf = plot_confusion_matrix(cnf_matrix, classes=class_names, normalize=True, title=title,
                                           fig_dir=fig_dir + '/cf_since_trigger', name=name)
        images_cf.append(imageio.imread(figname_cf))

        figname_roc, roc_auc[days_since_trigger] = compute_multiclass_roc_auc(class_names,
                                                                              to_categorical(y_test_on_day_i,
                                                                                             num_classes=nclasses),
                                                                              y_pred_prob_on_day_i, name=name,
                                                                              fig_dir=fig_dir + '/roc_since_trigger',
                                                                              title=title)
        images_roc.append(imageio.imread(figname_roc))

        figname_pr, pr_auc[days_since_trigger] = compute_precision_recall(class_names, to_categorical(y_test_on_day_i,
                                                                                                      num_classes=nclasses),
                                                                          y_pred_prob_on_day_i, name=name,
                                                                          fig_dir=fig_dir + '/pr_since_trigger',
                                                                          title=title)
        images_pr.append(imageio.imread(figname_pr))

        wlogloss[days_since_trigger] = plasticc_log_loss(to_categorical(y_test_on_day_i, num_classes=nclasses),
                                                         y_pred_prob_on_day_i, relative_class_weights=WLOGLOSS_WEIGHTS)

        plt.close()

        objids_filename = os.path.join(fig_dir + '/truth_table_since_trigger', 'objids_{}.csv'.format(name))
        predicted_table_filename = os.path.join(fig_dir + '/truth_table_since_trigger',
                                                'predicted_prob_{}.csv'.format(name))
        truth_table_filename = os.path.join(fig_dir + '/truth_table_since_trigger', 'truth_table_{}.csv'.format(name))
        np.savetxt(objids_filename, objids_test, fmt='%s')
        np.savetxt(predicted_table_filename, y_pred_prob_on_day_i)
        np.savetxt(truth_table_filename, to_categorical(y_test_on_day_i, num_classes=nclasses))
        # out_objtable = os.path.join(fig_dir + '/truth_table_since_trigger', 'out_obj_table_{}.txt'.format(name))
        # out_truth = os.path.join(fig_dir + '/truth_table_since_trigger', 'out_truth_{}.csv'.format(name))
        # out_pred = os.path.join(fig_dir + '/truth_table_since_trigger', 'out_pred_{}.csv'.format(name))
        # make_tables(objids_filename, predicted_table_filename, truth_table_filename, directory='', processes=2, out_objtable=out_objtable, out_truth=out_truth, out_pred=out_pred)

    imageio.mimsave(os.path.join(fig_dir, 'animation_cf_since_trigger.gif'), images_cf, duration=0.25)
    imageio.mimsave(os.path.join(fig_dir, 'animation_roc_since_trigger.gif'), images_roc, duration=0.25)
    imageio.mimsave(os.path.join(fig_dir, 'animation_pr_since_trigger.gif'), images_pr, duration=0.25)

    font = {'family': 'normal',
            'size': 36}

    matplotlib.rc('font', **font)

    # Plot ROC AUC vs time
    plt.figure("ROC AUC vs time", figsize=(13, 12))
    roc_auc = pd.DataFrame(roc_auc).transpose()
    names = list(roc_auc.keys())

    plt.plot(list(roc_auc['micro'].index), list(roc_auc['micro'].values), color='navy', linestyle=':', linewidth=4,
             label='micro-average')
    for classnum, classname in enumerate(class_names):
        if classname not in names or classnum == 0:
            print(classname)
            continue
        times = list(roc_auc[classname].index)
        aucs = list(roc_auc[classname].values)
        plt.plot(times, aucs, '-', label=classname, color=COLORS[classnum], linewidth=4)

    plt.legend(frameon=True, fontsize=25)
    plt.xlabel("Days since trigger (rest frame)")  # , fontsize=18)
    plt.ylabel("AUC")  # , fontsize=15)
    plt.tight_layout()
    plt.savefig(os.path.join(fig_dir, "auc_roc_vs_time.pdf"))
    plt.close()

    # # Plot Precision AUC vs time
    # plt.figure("ROC AUC vs time", figsize=(12, 11))
    # pr_auc = pd.DataFrame(pr_auc).transpose()
    # names = list(pr_auc.keys())
    # plt.plot(list(pr_auc['micro'].index), list(pr_auc['micro'].values), color='navy', linestyle=':', linewidth=4, label='micro')
    # for classnum, classname in enumerate(class_names):
    #     if classname not in names or classnum == 0:
    #         print(classname)
    #         continue
    #     times = list(pr_auc[classname].index)
    #     aucs = list(pr_auc[classname].values)
    #     plt.plot(times, aucs, '-', label=classname, color=COLORS[classnum], linewidth=2)
    #
    # plt.legend(frameon=False, fontsize=19)
    # plt.xlabel("Days since trigger")  # , fontsize=18)
    # plt.ylabel("AUC")  # , fontsize=15)
    # plt.savefig(os.path.join(fig_dir, "auc_pr_vs_time.pdf"))
    # plt.close()

    font = {'family': 'normal',
            'size': 36}
    matplotlib.rc('font', **font)

    # Plot weighted log loss vs time
    plt.figure("Weighted log loss vs time", figsize=(13, 12))
    plt.plot(list(wlogloss.keys()), list(wlogloss.values()), linewidth=4, label='Weighted Log loss')
    plt.xlabel("Days since trigger (rest frame)")  # , fontsize=18)
    plt.ylabel("Weighted log loss")  # , fontsize=15)
    plt.tight_layout()
    plt.savefig(os.path.join(fig_dir, "wlogloss_vs_time.pdf"))
    plt.close()
    print(wlogloss)




# -*- coding: UTF-8 -*-
'''
Feature definitions for ANTARES
'''

from __future__ import unicode_literals
__author__     = 'Gautham Narayan'
__email__      = 'gsnarayan@gmail.com'
__maintainer__ = 'Gautham Narayan'
__license__    = 'GPLv3'
__version__    = '0.1'
__status__     = 'alpha'



# -*- coding: UTF-8 -*-
"""
Functions for the derivation of base features with LAobjects
"""

from __future__ import absolute_import
from __future__ import unicode_literals
import numpy as np


class PlasticcMixin(object):
    """
    Methods specific to the Phootometric LSST Astronomical Time-series
    Classification Challenge (PLASTiCC)
    """
    def finalize_plasticc(self):
        if not 'photflag' in self._extra_cols:
            message = 'Object {} does not appear to be a valid PLAsTiCC LC. Must have `photflag`'.format(self.objectId)
            raise RuntimeError(message)

        trigger = (self.photflag == 6144)
        if len(self.time[trigger]) != 1:
            message = 'Object {} does not have a unique trigger'.format(self.objectId)
            raise RuntimeError(message)

        time_trigger   = self.time[trigger][0]
        deltaT_trigger = np.abs(self.time - time_trigger)

        for pb in self.filters:
            # find observations in each passband that are non-detections or trigger with time to trigger < season length
            indf = np.where((self.passband == pb) & ((self.photflag == 0) | (self.photflag == 6144)) & (deltaT_trigger < 90))
            if len(deltaT_trigger[indf]) > 0:
                best_ref = deltaT_trigger[indf].argmin()
            else:
                indf = np.where((self.passband == pb) & (self.photflag == 4096) & (deltaT_trigger < 90))
                if len(deltaT_trigger[indf]) > 0:
                    best_ref = deltaT_trigger[indf].argmin()
                else:
                    continue
            pb_ref = self.flux[indf][best_ref]
            key='{}Ref'.format(pb)
            self.header[key] = pb_ref 



import numpy as np
from astrorapid.classify import Classify


def rapid_stage(locus_data):
    locus_properties = locus_data.get_properties()
    objid = locus_properties['alert_id']
    ra = locus_properties['ra']
    dec = locus_properties['dec']
    redshift = 0.
    mwebv = 0.2

    alert_id, mjd, ras, decs, passband, mag, magerr, zeropoint, = locus_data.get_time_series('ra', 'dec', 'ztf_fid', 'ztf_magpsf', 'ztf_sigmapsf', 'ztf_magzpsci')

    flux = 10. ** (-0.4 * (mag - zeropoint))
    fluxerr = np.abs(flux * magerr * (np.log(10.) / 2.5))
    passband = np.where(passband == 1, 'g', passband)
    passband = np.where(passband == 2, 'r', passband)
    photflag = [0] * int(len(mjd) / 2 - 3) + [6144] + [4096] * int(len(mjd) / 2 + 2)

    print(locus_data.get_time_series('ra', 'dec', 'ztf_fid', 'ztf_magpsf', 'ztf_sigmapsf', 'ztf_magzpsci', 'ztf_diffmaglim'))

    light_curve_list = [(mjd, flux, fluxerr, passband, zeropoint, photflag, ra, dec, objid, redshift, mwebv)]

    classification = Classify(light_curve_list)
    predictions = classification.get_predictions()
    print(predictions)

    # classification.plot_light_curves_and_classifications()
    # classification.plot_classification_animation()



import numpy as np


def calc_luminosity(flux, fluxerr, mu):
    """ Normalise flux light curves with distance modulus.

    Parameters
    ----------
    flux : array
        List of floating point flux values.
    fluxerr : array
        List of floating point flux errors.
    mu : float
        Distance modulus from luminosity distance.

    Returns
    -------
    fluxout : array
        Same shape as input flux.
    fluxerrout : array
        Same shape as input fluxerr.

    """

    d = 10 ** (mu/5 + 1)
    dsquared = d**2

    norm = 1e18

    fluxout = flux * (4 * np.pi * dsquared/norm)
    fluxerrout = fluxerr * (4 * np.pi * dsquared/norm)

    return fluxout, fluxerrout


def aggregate_sntypes(reverse=False):
    if reverse:
        aggregate_map = {99: (45, 61, 62, 63, 90, 92),
                         1: (11, 1),
                         5: (3,13),
                         6: (2, 12, 14),
                         41: (41,),
                         43: (43,),
                         50: (50,),
                         51: (51,),
                         60: (60,),
                         64: (64,),
                         70: (70,),
                         80: (80,),
                         81: (81,),
                         83: (83,),
                         91: (91,),
                         93: (93,)}
    else:
        aggregate_map = {1: 1,
                         11: 1,
                         2: 6,
                         3: 5,
                         12: 6,
                         13: 5,
                         14: 6,
                         41: 41,
                         43: 43,
                         45: 45,
                         50: 50,
                         51: 51,
                         60: 60,
                         61: 61,
                         62: 62,
                         63: 63,
                         64: 64,
                         70: 70,
                         80: 80,
                         81: 81,
                         83: 83,
                         90: 90,
                         91: 91,
                         92: 92,
                         93: 93,
                         }

    return aggregate_map



import os
from keras.models import Sequential
from keras.models import load_model
from keras.layers import Dense, Input
from keras.layers import LSTM, GRU
from keras.layers import Dropout, BatchNormalization, Activation, TimeDistributed
from keras.layers.convolutional import Conv1D, Conv2D
from keras.layers.convolutional import MaxPooling1D, MaxPooling2D

from astrorapid.prepare_arrays import PrepareTrainingSetArrays


def train_model(X_train, X_test, y_train, y_test, sample_weights=None, fig_dir='.', retrain=True):
    model_filename = os.path.join(fig_dir, "keras_model.hdf5")

    if not retrain and os.path.isfile(model_filename):
        model = load_model(model_filename)
    else:
        num_classes = y_test.shape[-1]

        model = Sequential()

        # model.add(Conv1D(filters=32, kernel_size=3))
        # model.add(BatchNormalization())
        # model.add(Activation('relu'))
        # model.add(MaxPooling1D(pool_size=1))
        # model.add(Dropout(0.2, seed=42))

        model.add(GRU(100, return_sequences=True))
        model.add(Dropout(0.2, seed=42))
        model.add(BatchNormalization())

        model.add(GRU(100, return_sequences=True))
        model.add(BatchNormalization())
        model.add(Dropout(0.2, seed=42))

        model.add(TimeDistributed(Dense(num_classes, activation='softmax')))
        model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])
        model.fit(X_train, y_train, validation_data=(X_test, y_test), epochs=25, batch_size=64, verbose=2, sample_weight=sample_weights)

        print(model.summary())
        model.save(model_filename)

    return model


def main():
    passbands = ('g', 'r')
    contextual_info = ()

    aggregate_classes = True
    reread_hdf5_data = False
    retrain_rnn = False

    otherchange = ''
    nchunks = 1000

    # Train + Test cuts
    zcut = 0.5
    bcut = True
    variablescut = True

    training_set_dir = 'training_set_files'
    if not os.path.exists(training_set_dir):
        os.makedirs(training_set_dir)


    fpath = '/Users/danmuth/PycharmProjects/earlyclass/lc_data_MSIP_ZTF_20180716_full_lc_z0.5_mr23_ebv02_widercuts5.hdf5'

    fig_dir = os.path.join(training_set_dir, 'Figures', 'classify', 'ZTF_25epochs{}ag{}_ci{}_fp{}_zcut{}_bcut{}_varcut{}'.format(otherchange, aggregate_classes, contextual_info, os.path.basename(fpath), zcut, bcut, variablescut))
    for dirname in [fig_dir, fig_dir+'/cf_since_trigger', fig_dir+'/cf_since_t0', fig_dir+'/roc_since_trigger', fig_dir+'/lc_pred', fig_dir+'/pr_since_trigger', fig_dir+'/truth_table_since_trigger']:
        if not os.path.exists(dirname):
            os.makedirs(dirname)

    preparearrays = PrepareTrainingSetArrays(passbands, contextual_info, reread_hdf5_data, aggregate_classes, bcut, zcut, variablescut, nchunks=10000)
    X_train, X_test, y_train, y_test, labels_train, labels_test, class_names, class_weights, sample_weights, timesX_train, timesX_test, orig_lc_train, orig_lc_test, objids_train, objids_test = preparearrays.prepare_training_set_arrays(fpath, otherchange)
    model = train_model(X_train, X_test, y_train, y_test, sample_weights=sample_weights, fig_dir=fig_dir, retrain=retrain_rnn)
    plot_metrics(class_names, model, X_test, y_test, fig_dir, timesX_test=timesX_test, orig_lc_test=orig_lc_test, objids_test=objids_test, passbands=passbands)


if __name__ == '__main__':
    main()



"""
Example usage:
nice -n 19 python -m astrorapid.read_from_database.read_light_curves_from_database --offset 0 --offsetnext 1000 --nprocesses 8 --savename 'testing'
"""

import os
import numpy as np
import h5py
import multiprocessing as mp
import pandas as pd
import argparse

from astrorapid.read_from_database.get_data import GetData
from astrorapid.process_light_curves import InputLightCurve


def read_light_curves_from_sql_database(data_release, fname, field_in='%', model_in='%', batch_size=100, offset=0, sort=True, redo=False, passbands=('g', 'r'), known_redshift=True):
    print(fname)

    extrasql = ''  # "AND (objid LIKE '%00' OR objid LIKE '%50' OR sim_type_index IN (51,61,62,63,64,84,90,91,93))"  # ''#AND sim_redshift_host < 0.5 AND sim_peakmag_r < 23'
    getter = GetData(data_release)
    result = getter.get_lcs_data(columns=['objid', 'ptrobs_min', 'ptrobs_max', 'sim_peakmag_r', 'sim_redshift_host', 'mwebv', 'sim_dlmu', 'peakmjd', 'mwebv', 'ra', 'decl', 'hostgal_photoz', 'hostgal_photoz_err'],
                                 field=field_in, model=model_in, snid='%', limit=batch_size, offset=offset, shuffle=False, sort=sort, extrasql=extrasql)

    store = pd.HDFStore(fname)

    for head, phot in result:
        objid, ptrobs_min, ptrobs_max, peakmag, redshift, mwebv, dlmu, peakmjd, mwebv, ra, dec, photoz, photozerr = head

        field, model, base, snid = objid.split('_')

        lc = getter.convert_pandas_lc_to_recarray_lc(phot, passbands=passbands)

        inputlightcurve = InputLightCurve(lc['mjd'], lc['flux'], lc['dflux'], lc['pb'], lc['zpt'], lc['photflag'], ra,
                                          dec, objid, redshift, mwebv, known_redshift=known_redshift,
                                          training_set_parameters={'class_number': int(model), 'peakmjd': peakmjd})

        savepd = inputlightcurve.preprocess_light_curve()
        store.append(objid, savepd)

    store.close()
    print("saved %s" % fname)


def combine_hdf_files(save_dir, combined_savename, training_set_dir):
    fnames = os.listdir(save_dir)
    fname_out = os.path.join(training_set_dir, 'earlyclass', combined_savename)
    output_file = h5py.File(fname_out, 'w')


    for n, f in enumerate(fnames):
        print(n, f)
        try:
            f_hdf = h5py.File(os.path.join(save_dir, f), 'r')
            for objid in f_hdf.keys():
                objid = objid.encode('utf-8')
                h5py.h5o.copy(f_hdf.id, objid, output_file.id, objid)
            f_hdf.close()
        except OSError as e:
            print("Failed to open file", "f")
            print(e)
    output_file.close()


def create_all_hdf_files(args):
    data_release, i, save_dir, field_in, model_in, batch_size, sort, redo, passbands = args
    offset = batch_size * i
    fname = os.path.join(save_dir, 'earlylc_{}.hdf5'.format(i))
    read_light_curves_from_sql_database(data_release=data_release, fname=fname, field_in=field_in, model_in=model_in,
                           batch_size=batch_size, offset=offset, sort=sort, redo=redo, passbands=passbands)


def main():
    passbands = ('g', 'r')
    data_release = 'ZTF_20180716'
    field = 'MSIP'
    model = '%'

    # Get number of objects
    # extrasql = ''
    # getter = GetData(data_release)
    # nobjects = next(getter.get_lcs_headers(field=field, model=model, get_num_lightcurves=True, big=False, extrasql=extrasql))
    # print("{} objects for model {} in field {}".format(nobjects, model, field))

    batch_size = 1000
    sort = True
    redo = True

    parser = argparse.ArgumentParser()
    parser.add_argument('-o', "--offset", type=int)
    parser.add_argument('-n', "--offsetnext", type=int)
    parser.add_argument('-m', "--nprocesses", type=int, help='Number of multiprocessing processes. Default is 1.')
    parser.add_argument("--savename", type=str)
    args = parser.parse_args()
    if args.offset is not None:
        offset = args.offset
    else:
        offset = 0
    if args.offsetnext is not None:
        offset_next = args.offsetnext
    else:
        offset_next = 2200
    if args.nprocesses is not None:
        nprocesses = args.nprocesses
    else:
        nprocesses = 1
    if args.savename is not None:
        savename = args.savename
    else:
        savename = ""
    print(offset, offset_next)

    training_set_dir = 'training_set_files'
    save_dir = os.path.join(training_set_dir, 'saved_lc_{}_{}_{}'.format(field, data_release, savename))
    if not os.path.exists(save_dir) and offset == 0:
        os.makedirs(save_dir)

    # Multiprocessing
    i_list = np.arange(offset, offset_next)
    print(i_list)
    args_list = []
    file_list = os.listdir(save_dir)
    for i in i_list:
        if 'earlylc_{}.hdf5'.format(i) not in file_list:
            print(os.path.join(save_dir, 'earlylc_{}.hdf5'.format(i)))
            args_list.append((data_release, i, save_dir, field, model, batch_size, sort, redo, passbands))

    if nprocesses == 1:
        for args in args_list:
            create_all_hdf_files(args)
    else:
        pool = mp.Pool(nprocesses)
        results = pool.map_async(create_all_hdf_files, args_list)
        pool.close()
        pool.join()

    combine_hdf_files(save_dir, 'saved_lc_{}_{}_{}.hdf5'.format(field, data_release, savename), training_set_dir)


if __name__ == '__main__':
    main()




# -*- coding: UTF-8 -*-
"""
Functions for the derivation of base features with LAobjects
"""

from __future__ import absolute_import
from __future__ import unicode_literals
import numpy as np
import pandas as pd
import math
from six.moves import zip
import scipy.stats
import scipy.interpolate as scinterp
from astropy.stats import median_absolute_deviation
from cesium import featurize
from .. import constants
from . import stats_computation


class BaseMixin(object):
    """
    Methods to derive baseline features for LAobjects
    """

    def get_lc(self, recompute=False):
        """
        Most often, we want a lightcurve broken up passband by passband to
        compute features. This stores the input light curve in that format.
        Return a lightcurve suitable for computing features
        """
        outlc = getattr(self, '_outlc', None)
        if outlc is None or recompute:
            filters = self.filters
            outlc = {}

            # store the indices of each filter
            for i, pb in enumerate(self.filters):
                mask = np.where(self.passband == pb)[0]
                m2 = self.time[mask].argsort()
                ind = mask[m2]
                if len(ind) > 0:
                    outlc[pb] = ind
            self._outlc = outlc

        # for each of the default_cols, get the elements for t passband
        out = {}
        for pb, ind in outlc.items():
            out[pb] = [getattr(self, column)[ind] for column in self._default_cols]

        return out

    def get_amplitude(self, recompute=False):
        """
        Return the amplitude
        """
        outamp = getattr(self, 'amplitude', None)
        if outamp is not None:
            if not recompute:
                return outamp

        outamp = {}
        outlc = self.get_lc(recompute=recompute)

        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc

            newflux = tFluxRenorm
            if len(newflux) <= 1:  # if t Flux is empty or there's only measure
                amp = 0.
            else:
                f_99 = np.percentile(newflux, 99)
                f_01 = np.percentile(newflux, 1)
                f_50 = np.percentile(newflux, 50)
                amp = np.abs((f_99 - f_01) / (f_50 - f_01))
                if not np.isfinite(amp):
                    amp = 0.
            outamp[pb] = amp

        self.setattr_from_dict_default('amplitude', outamp, 0.)
        self.amplitude = outamp
        return outamp

    def get_filtered_amplitude(self, recompute=False):
        """
        Return the amplitude
        """
        outamp = getattr(self, 'filtered_amplitude', None)
        if outamp is not None:
            if not recompute:
                return outamp

        outamp = {}
        outlc = self.get_lc(recompute=recompute)

        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc

            photmask = tphotflag >= constants.GOOD_PHOTFLAG
            newflux = tFluxRenorm[photmask]

            if len(newflux) <= 1:  # if t Flux is empty or there's only measure
                amp = 0.
            else:
                f_99 = np.percentile(newflux, 99)
                f_01 = np.percentile(newflux, 1)
                f_50 = np.percentile(newflux, 50)
                amp = np.abs((f_99 - f_01) / (f_50 - f_01))
                if not np.isfinite(amp):
                    amp = 0.
            outamp[pb] = amp

        self.setattr_from_dict_default('filtered_amplitude', outamp, 0.)
        self.filtered_amplitude = outamp
        return outamp

    def get_stats(self, recompute=False):
        """
        Basic statistics for LAobject
        # min, max, mean, std, kurtosis, skewness
        """

        outstats = getattr(self, 'stats', None)
        if outstats is not None:
            if not recompute:
                return outstats

        outstats = {}
        outlc = self.get_lc(recompute=recompute)

        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc

            if len(tFluxRenorm) == 0:
                tstat = scipy.stats.describe([0., 0., 0.])
            else:
                tstat = scipy.stats.describe(tFluxRenorm)

            outstats[pb] = tstat

        self.stats = outstats
        return outstats

    def get_filtered_stats(self, recompute=False):
        """
        Basic statistics for LAobject from filtered observations
        # min, max, mean, std, kurtosis, skewness
        """

        outstats = getattr(self, 'filtered_stats', None)
        if outstats is not None:
            if not recompute:
                return outstats

        outstats = {}
        outlc = self.get_lc(recompute=recompute)

        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc
            mask = tphotflag >= constants.GOOD_PHOTFLAG
            tFluxRenorm = tFluxRenorm[mask]

            if len(tFluxRenorm) == 0:
                tstat = scipy.stats.describe([0., 0., 0.])
            else:
                tstat = scipy.stats.describe(tFluxRenorm)

            outstats[pb] = tstat

        self.filtered_stats = outstats
        return outstats

    def get_skew(self, recompute=False):
        """
        Different definition of skewness
        """
        outskew = getattr(self, 'skew', None)
        if outskew is not None:
            if not recompute:
                return outskew

        outskew = {}
        outlc = self.get_lc(recompute=recompute)
        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc
            npb = len(tFluxRenorm)
            tmean = np.mean(tFluxRenorm)
            tstd = np.std(tFluxRenorm)

            tskew = (1. / npb) * math.fsum(((tFluxRenorm - tmean) ** 3.) / (tstd ** 3.))
            outskew[pb] = tskew

        self.skew = outskew
        return outskew

    def get_StdOverMean(self, recompute=False):
        """
        Return the Standard Deviation over the Mean
        """
        outstats = self.get_stats(recompute=recompute)
        outSOMean = {pb: (x[3] ** 0.5 / x[2]) for pb, x in outstats.items()}
        return outSOMean

    def get_ShapiroWilk(self, recompute=False):
        """
        Get the Shapriro-Wilk W statistic and the p-value for the test
        """

        sw = getattr(self, 'ShapiroWilk', None)
        if sw is not None:
            if not recompute:
                return sw
        sw = {}
        outlc = self.get_lc(recompute=recompute)
        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc
            if len(tFlux) <= 3:
                continue
            tsw, tswp = scipy.stats.shapiro(tFluxRenorm)
            sw[pb] = (tsw, tswp)
        self.ShapiroWilk = sw
        return sw

    def get_Q31(self, recompute=False):
        """
        Get the Q31 of the lightcurve
        """
        q31 = getattr(self, 'Q31', None)
        if q31 is not None:
            if not recompute:
                return q31

        q31 = {}
        outlc = self.get_lc(recompute=recompute)

        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc
            tq31 = np.percentile(tFluxRenorm, 75) - np.percentile(tFluxRenorm, 25)
            q31[pb] = tq31
        self.Q31 = q31
        return q31

    def get_RMS(self, recompute=False):
        """
        Get the RMS of the lightcurve
        """
        rms = getattr(self, 'RMS', None)
        if rms is not None:
            if not recompute:
                return rms

        rms = {}
        outlc = self.get_lc(recompute=recompute)
        outstats = self.get_stats(recompute=recompute)

        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc

            photmask = tphotflag >= constants.GOOD_PHOTFLAG
            tFluxRenorm = tFluxRenorm[photmask]
            tFluxErrRenorm = tFluxErrRenorm[photmask]

            if len(tFluxRenorm) <= 1:  # if t Flux is empty or has only one element
                rms[pb] = 0.
                continue

            tmean = np.mean(tFluxRenorm)
            trms = math.fsum(((tFluxRenorm - tmean) / tFluxErrRenorm) ** 2.)
            trms /= math.fsum(1. / tFluxErrRenorm ** 2.)
            trms = trms ** 0.5
            rms[pb] = trms
        self.RMS = rms
        return rms

    def get_ShannonEntropy(self, recompute=False):
        """
        Compute the Shannon Entropy of the lightcurve distribution
        """

        entropy = getattr(self, 'entropy', None)
        if entropy is not None:
            if not recompute:
                return entropy
        entropy = {}
        outlc = self.get_lc(recompute=recompute)

        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc
            tEntropy = stats_computation.shannon_entropy(tFluxRenorm, tFluxErrRenorm)
            entropy[pb] = tEntropy
        self.entropy = entropy
        return entropy

    def get_MAD(self, recompute=False):
        """
        Compute the median absolute deviation of the lightcurve
        """

        mad = getattr(self, 'MAD', None)
        if mad is not None:
            if not recompute:
                return mad

        mad = {}
        outlc = self.get_lc(recompute=recompute)

        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc

            photmask = tphotflag >= constants.GOOD_PHOTFLAG
            tFluxRenorm = tFluxRenorm[photmask]
            if len(tFluxRenorm) == 0:  # if t Flux is empty
                tmad = 0.
            else:
                tmad = median_absolute_deviation(tFluxRenorm)
            mad[pb] = tmad
        self.MAD = mad
        return mad

    def get_vonNeumannRatio(self, recompute=False):
        """
        Compute the Von-Neumann Ratio of the lightcurve
        This is sometimes just called Eta in the context of variables

        The von Neumann ratio η was defined in 1941 by John von Neumann and serves as
        the mean square successive difference divided by the sample variance. When t
        ratio is small, it is an indication of a strong positive correlation between
        the successive photometric data points.  See: (J. Von Neumann, The Annals of
        Mathematical Statistics 12, 367 (1941))

        This seems like something that'd be much useful to compute from a phase
        curve...
        """

        vnr = getattr(self, 'VNR', None)
        if vnr is not None:
            if not recompute:
                return vnr

        vnr = {}
        outlc = self.get_lc(recompute=recompute)
        outstats = self.get_stats(recompute=recompute)

        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc
            delta = math.fsum(((tFluxRenorm[1:] - tFluxRenorm[:-1]) ** 2.) / (len(tFluxRenorm) - 1))
            tstats = outstats.get(pb)
            if tstats is None:
                continue
            tvar = tstats[3]
            tvnr = delta / (tvar)
            vnr[pb] = tvnr
        self.VNR = vnr
        return vnr

    def get_StetsonJ(self, recompute=False):
        """
        Compute the Stetson J statistic of the lightcurve
        http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.257.466&rep=rep1&type=pdf
        """

        stetsonJ = getattr(self, 'stetsonJ', None)
        if stetsonJ is not None:
            if not recompute:
                return stetsonJ

        stetsonJ = {}
        outlc = self.get_lc(recompute=recompute)
        outstats = self.get_stats(recompute=recompute)

        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc

            tstats = outstats.get(pb)
            if tstats is None:
                continue

            photmask = tphotflag >= constants.GOOD_PHOTFLAG
            tFluxRenorm = tFluxRenorm[photmask]
            tFluxErrRenorm = tFluxErrRenorm[photmask]

            tmean = tstats[2]
            npb = len(tFluxRenorm)

            if npb < 2:
                stetsonJ[pb] = 0.
                continue

            delta = (npb / (npb - 1)) * ((tFluxRenorm - tmean) / tFluxErrRenorm)
            val = np.nan_to_num(delta[0:-1] * delta[1:])
            sign = np.sign(val)
            tJ = math.fsum(sign * (np.abs(val) ** 0.5))
            stetsonJ[pb] = tJ
        self.stetsonJ = stetsonJ
        return stetsonJ

    def get_StetsonK(self, recompute=False):
        """
        Compute the Stetson K statistic of the lightcurve
        """

        stetsonK = getattr(self, 'stetsonK', None)
        if stetsonK is not None:
            if not recompute:
                return stetsonK

        stetsonK = {}
        outlc = self.get_lc(recompute=recompute)
        outstats = self.get_stats(recompute=recompute)

        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc

            tstats = outstats.get(pb)
            if tstats is None:
                stetsonK[pb] = 0.
                continue

            tmean = tstats[2]
            npb = len(tFluxRenorm)

            if npb < 2:
                stetsonK[pb] = 0.
                continue

            delta = (npb / (npb - 1)) * ((tFluxRenorm - tmean) / tFluxErrRenorm)
            tK = (np.sum(np.fabs(delta)) / npb) / np.sqrt(np.sum(delta * delta) / npb)
            tK = np.nan_to_num(tK)
            stetsonK[pb] = tK
        self.stetsonK = stetsonK
        return stetsonK

    def get_StetsonL(self, recompute=False):
        """
        Get the Stetson L variability index
        """
        stetsonL = getattr(self, 'stetsonL', None)
        if stetsonL is not None:
            if not recompute:
                return stetsonL

        stetsonJ = self.get_StetsonJ(recompute=recompute)
        stetsonK = self.get_StetsonK(recompute=recompute)

        stetsonL = {}
        for pb in stetsonJ:
            tJ = stetsonJ.get(pb, 0.)
            tK = stetsonK.get(pb, 0.)
            tL = tJ * tK / 0.798
            tL = np.nan_to_num(tL)
            stetsonL[pb] = tL
        self.stetsonL = stetsonL
        return stetsonL

    def get_AcorrIntegral(self, recompute=False):
        """
        Compute the Autocorrelation get_AcorrIntegral
        """
        AcorrInt = getattr(self, 'AcorrInt', None)
        if AcorrInt is not None:
            if not recompute:
                return AcorrInt

        AcorrInt = {}
        outlc = self.get_lc(recompute=recompute)
        outstats = self.get_stats(recompute=recompute)
        outrms = self.get_RMS(recompute=recompute)

        for j, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc

            tstats = outstats.get(pb)
            if tstats is None:
                continue
            tmean = tstats[2]

            trms = outrms[pb]
            if trms is None:
                continue
            if trms == 0.:
                AcorrInt[pb] = 0.
                continue

            npb = len(tFluxRenorm)
            t = np.arange(1, npb)
            sum_list = []
            val_list = []
            for i in t:
                sum_list.append(math.fsum((tFluxRenorm[0:npb - i] - tmean) * (tFluxRenorm[i:npb] - tmean)))
                val_list.append(1. / ((npb - i) * trms ** 2.))
            tAcorr = np.abs(math.fsum([x * y for x, y in zip(sum_list, val_list)]))
            AcorrInt[pb] = tAcorr
        self.AcorrInt = AcorrInt
        return AcorrInt

    def get_hlratio(self, recompute=False):
        """
        Compute the ratio of amplitude of observations higher than the average
        to those lower than the average, taking into account observed
        uncertainties. This ratio should be higher for eclipsing binaries than
        pulsating variables.
        """
        hlratio = getattr(self, 'hlratio', None)
        if hlratio is not None:
            if not recompute:
                return hlratio

        hlratio = {}
        outlc = self.get_lc(recompute=recompute)
        outstats = self.get_stats(recompute=recompute)

        for j, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc

            tWeight = 1. / tFluxErrRenorm

            tstats = outstats.get(pb)
            if tstats is None:
                continue
            tmean = tstats[2]
            il = tFluxRenorm > tmean
            wl = tWeight[il]
            wlsum = np.sum(wl)
            fl = tFluxRenorm[il]
            wl_weighted_std = np.sum(wl * (fl - tmean) ** 2) / wlsum

            ih = tFluxRenorm <= tmean
            wh = tWeight[ih]
            whsum = np.sum(wh)
            fh = tFluxRenorm[ih]
            wh_weighted_std = np.sum(wh * (fh - tmean) ** 2) / whsum

            hlratio[pb] = np.nan_to_num(np.sqrt(wl_weighted_std / wh_weighted_std))
        self.hlratio = hlratio
        return hlratio

    def get_color_amplitudes(self, recompute=False, passbands=('u', 'g', 'r', 'i', 'z', 'Y')):
        """
        Get the amplitude difference between passbands
        passbands = ('u', 'g', 'r', 'i', 'z', 'Y') # order of filters matters as it must be 'u-g' rather than 'g-u'
        """
        colorAmp = getattr(self, 'colorAmp', None)
        if colorAmp is not None:
            if not recompute:
                return colorAmp

        colorAmp = {}
        amp = self.get_filtered_amplitude(recompute=recompute)

        for i, pb1 in enumerate(passbands):
            for j, pb2 in enumerate(passbands):
                if i < j:
                    color = pb1 + '-' + pb2
                    if pb1 not in amp.keys():
                        amp[pb1] = 0.
                    if pb2 not in amp.keys():
                        amp[pb2] = 0.
                    colorAmp[color] = amp[pb1] - amp[pb2]

        self.colorAmp = colorAmp
        return colorAmp

    def get_color_mean(self, recompute=False, passbands=('u', 'g', 'r', 'i', 'z', 'Y')):
        """
        passbands = ('u', 'g', 'r', 'i', 'z', 'Y') # order of filters matters as it must be 'u-g' rather than 'g-u'
        """
        colorMean = getattr(self, 'colorMean', None)
        if colorMean is not None:
            if not recompute:
                return colorMean

        colorMean = {}
        outlc = self.get_lc(recompute=recompute)

        tmean = {}
        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc
            tmean[pb] = np.mean(tFluxRenorm)

        for i, pb1 in enumerate(passbands):
            for j, pb2 in enumerate(passbands):
                if i < j:
                    color = pb1 + '-' + pb2
                    if pb1 not in tmean.keys() or pb2 not in tmean.keys():
                        colorMean[color] = 0.
                        continue
                    colorMean[color] = tmean[pb1] / tmean[pb2]

        self.colorMean = colorMean
        return colorMean

    def get_historic_color(self, recompute=False, passbands=('u', 'g', 'r', 'i', 'z', 'Y')):
        historicColors = getattr(self, 'historicColors', None)
        if historicColors is not None:
            if not recompute:
                return historicColors

        historicColors = {}
        outlc = self.get_lc(recompute=recompute)

        tmedianflux = {}
        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc

            photmask = tphotflag >= constants.GOOD_PHOTFLAG

            tmedianflux[pb] = np.median(tFluxUnred[photmask])

        for i, pb1 in enumerate(passbands):
            for j, pb2 in enumerate(passbands):
                if i < j:
                    color = pb1 + '-' + pb2
                    if pb1 not in tmedianflux.keys() or pb2 not in tmedianflux.keys():
                        tmedianflux[color] = 0.
                        continue
                    historicColors[color] = tmedianflux[pb1] / tmedianflux[pb2]

        self.historicColors = historicColors
        return historicColors


    def get_cesium_features(self, recompute=False):
        """
        Compute all relevant cesium features.
        """

        cesium = getattr(self, 'cesium', None)
        if cesium is not None:
            if not recompute:
                return cesium

        cesium = {}
        outlc = self.get_lc(recompute=recompute)

        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc

            photmask = tphotflag >= constants.GOOD_PHOTFLAG
            ttime = ttime[photmask]
            tFluxRenorm = tFluxRenorm[photmask]
            tFluxErrRenorm = tFluxErrRenorm[photmask]

            features_general = ['flux_percentile_ratio_mid20',
                                'flux_percentile_ratio_mid50',
                                'flux_percentile_ratio_mid65',
                                'flux_percentile_ratio_mid80',
                                'max_slope',
                                'maximum',
                                'median',
                                'median_absolute_deviation',
                                'percent_amplitude',
                                'period_fast',
                                'qso_log_chi2_qsonu',
                                'qso_log_chi2nuNULL_chi2nu',
                                'fold2P_slope_90percentile',
                                'freq1_amplitude1',
                                'freq1_amplitude2',
                                'freq1_amplitude3',
                                'freq1_amplitude4',
                                'freq1_freq',
                                'freq1_lambda',
                                'freq1_rel_phase2',
                                'freq1_rel_phase3',
                                'freq1_rel_phase4',
                                'freq1_signif',
                                'freq2_amplitude1',
                                'freq2_amplitude2',
                                'freq2_amplitude3',
                                'freq2_amplitude4',
                                'freq2_rel_phase2',
                                'freq2_rel_phase3',
                                'freq2_rel_phase4',
                                'freq3_amplitude1',
                                'freq3_amplitude2',
                                'freq3_amplitude3',
                                'freq3_amplitude4',
                                'freq3_rel_phase2',
                                'freq3_rel_phase3',
                                'freq3_rel_phase4',
                                'freq_amplitude_ratio_21',
                                'freq_amplitude_ratio_31',
                                'freq_n_alias',
                                'freq_signif_ratio_21',
                                'freq_signif_ratio_31',
                                'freq_varrat',
                                'freq_y_offset',
                                'medperc90_2p_p',
                                'p2p_scatter_pfold_over_mad',
                                'p2p_ssqr_diff_over_var',
                                'scatter_res_raw']

            if len(tFluxRenorm) <= 1:  # if t Flux is empty
                fset_cesium = pd.DataFrame({f: {'channel': {0: 0}} for f in features_general})
            else:
                fset_cesium = featurize.featurize_time_series(times=ttime,
                                                              values=tFluxRenorm,
                                                              errors=tFluxErrRenorm,
                                                              features_to_use=features_general)

            cesium[pb] = fset_cesium

        self.cesium = cesium
        return cesium

    def get_rise_time(self, recompute=False):
        """
        Compute the risetime of the light curve.
        """

        risetime = getattr(self, 'risetime', None)
        if risetime is not None:
            if not recompute:
                return risetime

        risetime = {}
        outlc = self.get_lc(recompute=recompute)

        for i, pb in enumerate(outlc):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc

            photmask = tphotflag >= constants.GOOD_PHOTFLAG
            ttime = ttime[photmask]
            tFluxRenorm = tFluxRenorm[photmask]

            # Assuming that ttime = 0 is the peak_mjd, then:

            if len(tFluxRenorm) <= 1:  # if t Flux is nearly empty
                trisetime = -1.
            else:
                trisetime = 0 - ttime[0]
            risetime[pb] = trisetime

        self.risetime = risetime
        return risetime

    def get_rise_rate(self, recompute=False):
        """
        Compute the rise rate (slope) of the light curve.
        """

        riserate = getattr(self, 'riserate', None)
        if riserate is not None:
            if not recompute:
                return riserate

        risetime = self.get_rise_time(recompute=recompute)
        amplitude = self.get_filtered_amplitude(recompute=recompute)

        riserate = {}
        for pb in risetime:
            trisetime = amplitude.get(pb, -1.)
            tamplitude = amplitude.get(pb, 0.)
            if trisetime == -1.:
                triserate = 0
            else:
                triserate = tamplitude / trisetime
            riserate[pb] = triserate
        self.riserate = riserate
        return riserate

# features_all_cesium = ['all_times_nhist_numpeaks',
#                        'all_times_nhist_peak1_bin',
#                        'all_times_nhist_peak2_bin',
#                        'all_times_nhist_peak3_bin',
#                        'all_times_nhist_peak4_bin',
#                        'all_times_nhist_peak_1_to_2',
#                        'all_times_nhist_peak_1_to_3',
#                        'all_times_nhist_peak_1_to_4',
#                        'all_times_nhist_peak_2_to_3',
#                        'all_times_nhist_peak_2_to_4',
#                        'all_times_nhist_peak_3_to_4',
#                        'all_times_nhist_peak_val',
#                        'avg_double_to_single_step',
#                        'avg_err',
#                        'avgt',
#                        'cad_probs_1',
#                        'cad_probs_10',
#                        'cad_probs_20',
#                        'cad_probs_30',
#                        'cad_probs_40',
#                        'cad_probs_50',
#                        'cad_probs_100',
#                        'cad_probs_500',
#                        'cad_probs_1000',
#                        'cad_probs_5000',
#                        'cad_probs_10000',
#                        'cad_probs_50000',
#                        'cad_probs_100000',
#                        'cad_probs_500000',
#                        'cad_probs_1000000',
#                        'cad_probs_5000000',
#                        'cad_probs_10000000',
#                        'cads_avg',
#                        'cads_med',
#                        'cads_std',
#                        'mean',
#                        'med_double_to_single_step',
#                        'med_err',
#                        'n_epochs',
#                        'std_double_to_single_step',
#                        'std_err',
#                        'total_time',
#                        'General',
#                        'amplitude',
#                        'flux_percentile_ratio_mid20',
#                        'flux_percentile_ratio_mid35',
#                        'flux_percentile_ratio_mid50',
#                        'flux_percentile_ratio_mid65',
#                        'flux_percentile_ratio_mid80',
#                        'max_slope',
#                        'maximum',
#                        'median',
#                        'median_absolute_deviation',
#                        'minimum',
#                        'percent_amplitude',
#                        'percent_beyond_1_std',
#                        'percent_close_to_median',
#                        'percent_difference_flux_percentile',
#                        'period_fast',
#                        'qso_log_chi2_qsonu',
#                        'qso_log_chi2nuNULL_chi2nu',
#                        'skew',
#                        'std',
#                        'stetson_j',
#                        'stetson_k',
#                        'weighted_average',
#                        'Lomb-Scargle (Periodic)',
#                        'fold2P_slope_10percentile',
#                        'fold2P_slope_90percentile',
#                        'freq1_amplitude1',
#                        'freq1_amplitude2',
#                        'freq1_amplitude3',
#                        'freq1_amplitude4',
#                        'freq1_freq',
#                        'freq1_lambda',
#                        'freq1_rel_phase2',
#                        'freq1_rel_phase3',
#                        'freq1_rel_phase4',
#                        'freq1_signif',
#                        'freq2_amplitude1',
#                        'freq2_amplitude2',
#                        'freq2_amplitude3',
#                        'freq2_amplitude4',
#                        'freq2_freq',
#                        'freq2_rel_phase2',
#                        'freq2_rel_phase3',
#                        'freq2_rel_phase4',
#                        'freq3_amplitude1',
#                        'freq3_amplitude2',
#                        'freq3_amplitude3',
#                        'freq3_amplitude4',
#                        'freq3_freq',
#                        'freq3_rel_phase2',
#                        'freq3_rel_phase3',
#                        'freq3_rel_phase4',
#                        'freq_amplitude_ratio_21',
#                        'freq_amplitude_ratio_31',
#                        'freq_frequency_ratio_21',
#                        'freq_frequency_ratio_31',
#                        'freq_model_max_delta_mags',
#                        'freq_model_min_delta_mags',
#                        'freq_model_phi1_phi2',
#                        'freq_n_alias',
#                        'freq_signif_ratio_21',
#                        'freq_signif_ratio_31',
#                        'freq_varrat',
#                        'freq_y_offset',
#                        'linear_trend',
#                        'medperc90_2p_p',
#                        'p2p_scatter_2praw',
#                        'p2p_scatter_over_mad',
#                        'p2p_scatter_pfold_over_mad',
#                        'p2p_ssqr_diff_over_var',
#                        'scatter_res_raw']



# -*- coding: UTF-8 -*-
"""
ANTARES Object class specification
"""

from __future__ import absolute_import
from __future__ import unicode_literals
import warnings
import numpy as np
from . import constants
from .features.base import BaseMixin
from astropy.stats import sigma_clip
import extinction

__all__ = ['LAobject']


class LAobject(BaseMixin):
    """
    ANTARES object - locus aggregated alert lightcurve and feature encapsulator

    :py:func:`antares.model.helper.MakeTouchstoneObject` creates a
    :py:class:`ANTARES_object` instance from an event alert.

    TODO: rename ``MakeTouchstoneObject`` something more useful

    Parameters
    ----------
    locusId : str or int
        the ANTARES locus aggegated alert ID
    objectId : str or int
        The unique identifier for each Object - it's name from the source survey e.g. diaObjectId
    time : array-like
        The mid-point of the observation - e.g. midPointTai from DIASource
    flux : array-like
        The calibrated flux - e.g. totFlux of each DIASource
    fluxErr : array-like
        The calibrated flux uncertainty - e.g. totFluxErr of each DIASource
    obsId : array-like
        The list of IDS  corresponding to the observations - e.g. diaSourceId
    photflag : array-like
        Photometry flag corresponding to each DIASource
        The flag above which the photometry is asserted to be good is defined
        in GOOD_PHOTFLAG is defined in `constants.py`
    passband : array-like
        The passband of each observation
    zeropoint : float or array-like
        The zeropoint of each observation - used to convert to Pogson magnitudes if possible
        The default zeropoint DEFAULT_ZEROPOINT is defined in `constants.py`
    per : bool
        Boolean flag indicating if this object is considered a periodic variable or not
    best_period : float or None
        If `per`, then the best-estimate of the period of this object
    header : None or dict
        A header entry as a dictionary of values that should be preserved during processing
    mag : bool
        Is this input light curve provided in magnitudes? Converts to flux
        using zeropoint if True
    preprocess : bool
        Clean up invalid values from the light curve
    clean : bool
        Attempt to remove outliers from the light curve. Implies preprocess. Use with caution.
    ebv : float
        Milky-Way extinction on line of sight to this object
        Assumed to be zero if not provided
        Applied to construct the unreddened flux - fluxUnred
    renorm : bool
        NOT IMPLMENTED, DEPRECATED
    mu : float
        distance modulus of this object - if renorm is True, the fluxes are
        corrected from this distance modulus to 10pc - note that there is no
        k-correction being applied.
        NOT IMPLMENTED, DEPRECATED

    Notes
    -----
        Follows the DPDD as of 20170810:
        https://docushare.lsstcorp.org/docushare/dsweb/Get/LSE-163/

        ``time``, ``flux``, ``fluxErr``, ``obsID``, ``passband`` must be 1-D
        arrays with the same shape, and have at least one element
    """

    def __init__(self, locusId, objectId, time, flux, fluxErr, obsId, photflag, passband,
                 zeropoint=constants.DEFAULT_ZEROPOINT,
                 per=False, best_period=None, header=None, mag=False, preprocess=True,
                 clean=False, renorm=False, ebv=0., mu=None, **kwargs):

        # object name or ID - an alert ID generated by ANTARES or pre-supplied
        # note that if ANTARES is generating its own IDs, then we need someway
        # to store a DIAObject ID

        # scalars
        self.objectId = objectId
        self.locusId = locusId
        self.ebv = float(ebv)

        # vectors
        self.time = np.array(time).astype('f')
        zeropoint = np.atleast_1d(zeropoint)
        # if there's a single number for zeropoint, make sure that it is an array
        if len(zeropoint) == 1:
            zeropoint = np.repeat(zeropoint, _tshape[0])
        self.zeropoint = zeropoint

        # convert input magnitudes to fluxes
        if mag:
            mag = np.array(flux).astype('f')
            magErr = np.array(fluxErr).astype('f')
            self.flux = 10. ** (-0.4 * (mag - zeropoint))
            self.fluxErr = np.abs(self.flux * magErr * (np.log(10.) / 2.5))
        else:
            self.flux = np.array(flux).astype('f')
            self.fluxErr = np.array(fluxErr).astype('f')
        self.obsId = np.array(obsId)
        self.passband = passband
        self.photflag = photflag

        # names of extra vectors
        self._extra_cols = []

        # check that the arrays have the same shape
        assert self.time.ndim == 1
        _tshape = self.time.shape
        assert _tshape == self.flux.shape
        assert _tshape == self.fluxErr.shape
        assert _tshape == self.obsId.shape
        assert _tshape == self.passband.shape
        assert _tshape == self.photflag.shape
        assert _tshape == self.zeropoint.shape

        for key, val in kwargs.items():
            if np.isscalar(val):
                setattr(self, key, val)
            else:
                valarr = np.array(val)
                setattr(self, key, valarr)
                if _tshape == valarr.shape:
                    self._extra_cols.append(key)

        # keep track of what passbands we have
        self.filters = set(self.passband)

        # save the other inputs
        self.per = per
        if self.per:
            self.best_period = best_period
            self.lcPeriodic = np.nan
            self.lcNonPeriodic = None
        else:
            self.best_period = None
            self.lcPeriodic = None
            self.lcNonPeriodic = np.nan

        self.trigger_time = self.time[self.photflag == constants.TRIGGER_PHOTFLAG][0]

        if header is None:
            header = {}
        self.header = header

        # pre-processing = remove invalid values
        # cleaning = remove outliers
        # cleaning the lightcurve implies pre-processing
        if clean and not preprocess:
            preprocess = True

        if preprocess:
            # exclude bad values
            mask = np.isfinite(self.fluxErr)  # np.logical_and((self.fluxErr > 1E-8), np.isfinite(self.fluxErr))  #
            mask = np.logical_and(mask, np.isfinite(self.flux))
            mask = np.logical_and(mask, np.isfinite(self.zeropoint))
            # mask = np.logical_and(mask, self.photflag == constants.BAD_PHOTFLAG)

            # these cuts are only applied if the light curve is provided in magnitudes
            # they handle annoying dummy values and very low S/N points
            if mag:
                mask = np.logical_and(mask, self.flux > 0.)
                mask = np.logical_and(mask, self.flux <= 99.)
                mask = np.logical_and(mask, self.fluxErr <= 9.)

            # if any of the filtered points are flagged as good detections with
            # PHOTFLAG > 0, save them irrespective
            saveind = (self.photflag[mask] >= constants.GOOD_PHOTFLAG)
            mask[saveind] = True

            # apply the mask
            self.time = self.time[mask]
            self.flux = self.flux[mask]
            self.fluxErr = self.fluxErr[mask]
            self.obsId = self.obsId[mask]
            self.photflag = self.photflag[mask]
            self.passband = self.passband[mask]
            self.zeropoint = self.zeropoint[mask]

            for key in self._extra_cols:
                val = getattr(self, key)
                setattr(self, key, val[mask])

            if not clean:
                return self.finalize()

            # we should still update filters in case all the data in a filter was invalid
            self.filters = set(self.passband)

            # begin cleaning the lightcurve - hic sunt dracones
            # remove points with S/N < 1
            mask = self.fluxErr < np.abs(self.flux)

            # apply the mask
            self.time = self.time[mask]
            self.flux = self.flux[mask]
            self.fluxErr = self.fluxErr[mask]
            self.obsId = self.obsId[mask]
            self.photflag = self.photflag[mask]
            self.passband = self.passband[mask]
            self.zeropoint = self.zeropoint[mask]
            for key in self._extra_cols:
                val = getattr(self, key)
                setattr(self, key, val[mask])

            t = None
            f = None
            df = None
            zpt = None
            pbs = None
            oid = None
            pf = None
            filtered_extra_cols = {}

            # do some sigmaclipping to reject outliers
            for pb in self.filters:
                indf = (self.passband == pb)
                filtered_err = sigma_clip(self.fluxErr[indf], sigma=3., iters=5, copy=True)
                bad1 = filtered_err.mask

                saveind = np.where(self.photflag[indf][bad1] >= constants.GOOD_PHOTFLAG)
                bad1[saveind] = False

                useind = ~bad1

                if t is None:
                    t = self.time[indf][useind]
                else:
                    t = np.concatenate((t, self.time[indf][useind]))

                if f is None:
                    f = self.flux[indf][useind]
                else:
                    f = np.concatenate((f, self.flux[indf][useind]))

                if df is None:
                    df = self.fluxErr[indf][useind]
                else:
                    df = np.concatenate((df, self.fluxErr[indf][useind]))

                if zpt is None:
                    zpt = self.zeropoint[indf][useind]
                else:
                    zpt = np.concatenate((zpt, self.zeropoint[indf][useind]))

                if pbs is None:
                    pbs = self.passband[indf][useind]
                else:
                    pbs = np.concatenate((pbs, self.passband[indf][useind]))

                if oid is None:
                    oid = self.obsId[indf][useind]
                else:
                    oid = np.concatenate((oid, self.obsId[indf][useind]))

                if pf is None:
                    pf = self.photflag[indf][useind]
                else:
                    pf = np.concatenate((pf, self.photflag[indf][useind]))

                for key in self._extra_cols:
                    orig = getattr(self, key)
                    val = filtered_extra_cols.get(key, None)
                    if val is None:
                        val = orig[indf][useind]
                        filtered_extra_cols[key] = val
                    else:
                        val = np.concatenate((val, orig[indf][useind]))
                        filtered_extra_cols[key] = val

            self.time = t
            self.flux = f
            self.fluxErr = df
            self.passband = pbs
            self.zeropoint = zpt
            self.obsId = oid
            self.photflag = photflag

            for key in self._extra_cols:
                val = filtered_extra_cols.get(key)
                setattr(self, key, val)

            # finalize the light curve after cleaning
            return self.finalize()

    def _remove_flux_extinction(self):
        """
        Remove extinction for light curve assuming Fitzpatrick '99 reddening
        law, given some value of E(B-V)
        """
        self.fluxUnred = self.flux.copy()
        self.fluxErrUnred = self.fluxErr.copy()
        self.fluxRenorm = self.flux.copy()
        self.fluxErrRenorm = self.fluxErr.copy()

        # Using negative a_v so that extinction.apply works in reverse and removes the extinction
        extinctions = extinction.fitzpatrick99(wave=self._good_filter_wave, \
                                               a_v=-3.1 * self.ebv, r_v=3.1, unit='aa')

        for i, pb in enumerate(self._good_filters):
            mask = (self.passband == pb)

            flux_pb = self.flux[mask]
            fluxerr_pb = self.fluxErr[mask]
            npbobs = len(flux_pb)

            if npbobs > 1:
                # there's at least enough observations to find minimum and maximum
                flux_out = extinction.apply(extinctions[i], flux_pb, inplace=False)
                fluxerr_out = extinction.apply(extinctions[i], fluxerr_pb, inplace=False)
                self.fluxUnred[mask] = flux_out
                self.fluxErrUnred[mask] = fluxerr_out

                minfluxpb = flux_out.min()
                maxfluxpb = flux_out.max()
                norm = maxfluxpb - minfluxpb

                self.fluxRenorm[mask] = flux_out - minfluxpb
                self.fluxErrRenorm[mask] = fluxerr_out

                self.fluxRenorm[mask] /= norm
                self.fluxErrRenorm[mask] /= norm
            elif npbobs == 1:
                # deal with the case with one observation in this passband by setting renorm = 0.5
                flux_out = extinction.apply(extinctions[i], flux_pb, inplace=False)
                fluxerr_out = extinction.apply(extinctions[i], fluxerr_pb, inplace=False)
                self.fluxUnred[mask] = flux_out
                self.fluxErrUnred[mask] = fluxerr_out

                norm = self.fluxUnred[mask] / 0.5
                self.fluxRenorm[mask] /= norm
                self.fluxErrRenorm[mask] /= norm
            else:
                pass

        self._default_cols = ['time', 'flux', 'fluxErr', 'fluxUnred', 'fluxErrUnred', \
                              'fluxRenorm', 'fluxErrRenorm', 'photflag', 'zeropoint', 'obsId']
        return

    def finalize(self):
        """
        Finalizes the LAobject by setting the list of available
        filters and the light curve length after filtering.

        HACK : the filters are limited to the PLAsTiCC ugrizY set since the
        PropertyTable SQL in ANTARES must have pre-defined keys
        """
        # this forces only some filters will be used for feature computation
        # this is not ideal, but a necessary stop-gap while we revise
        # the PropertyTable SQL
        self._good_filters = ['u', 'g', 'r', 'i', 'z', 'Y']
        self._good_filter_wave = np.array([3569.5, 4766.5, 6214.5, 7544.5, 8707.5, 10039.5])

        use_filters = set(self._good_filters) & self.filters
        if not self.filters.issubset(use_filters):
            message = 'Number of useful filters ({}) does not equal number available filters ({}) - some filters will not be used'.format(
                ''.join(use_filters), ''.join(self.filters))
            warnings.warn(message, RuntimeWarning)
        self.filters = set(use_filters)
        mask = np.array([True if x in self.filters else False for x in self.passband])

        if mask.size:  # Not empty arrays
            self.time = self.time[mask]
            self.flux = self.flux[mask]
            self.fluxErr = self.fluxErr[mask]
            self.obsId = self.obsId[mask]
            self.passband = self.passband[mask]
            self.zeropoint = self.zeropoint[mask]
            for key in self._extra_cols:
                val = getattr(self, key)
                setattr(self, key, val[mask])

        self.nobs = len(self.time)
        if self.nobs == 0:
            message = 'Object {} with locus ID {} has no good observations.'.format(self.objectId, self.locusId)
            raise ValueError(message)

        return self._remove_flux_extinction()

    def setattr_from_dict_default(self, rootname, values_dict, default_value):
        """
        Set attributes for the LAobject from a values dictionary indexed by passband
        The set attribute names are rootname_passband

        Attributes are only set for passbands listed in _good_filters i.e. if
        the light curve has passbands that were non-standard, then no features
        are set.

        If a passband listed in _good_filters is not present in the
        values_dict, the default_value is set
        """
        for pb in self._good_filters:
            setattr(self, '{}_{}'.format(rootname, pb), values_dict.get(pb, default_value))



# -*- coding: UTF-8 -*-
"""
Functions for parametric representations of LAobjects

"""

from __future__ import absolute_import
from __future__ import unicode_literals
import numpy as np
from .. import constants
from iminuit import Minuit


class ParametricMixin(object):

    def bazin_fit(self, minpbobs=6, recompute=False):
        """
        Fit a Bazin model to the LAobject
        """
        bazin_params = getattr(self, 'bazin_params', None)
        if bazin_params is not None:
            if not recompute:
                return bazin_params

        bazin_params = {}
        trigger_time = self.trigger_time

        filters = self.filters
        outlc = self.get_lc(recompute=recompute)

        for i, pb in enumerate(filters):
            tlc = outlc.get(pb)
            ttime, tFlux, tFluxErr, tFluxUnred, tFluxErrUnred, tFluxRenorm, tFluxErrRenorm, tphotflag, tzeropoint, tobsId = tlc

            # enough observations in this band to attempt to fit a model
            # there's 5 parameters, so there should be at least 6
            nobs = len(ttime)
            if nobs < minpbobs:
                print(nobs, "Not enough obs")
                continue

            # demand at least 3 points before or at trigger
            pre_trigger = ttime <= trigger_time 
            npre_trigger = len(ttime[pre_trigger])
            if npre_trigger <= 3:
                print(npre_trigger, "Not enough obs before trigger")
                continue

            # demand at least 3 points after trigger
            npost_trigger = nobs - npre_trigger 
            if npost_trigger <= 3:
                print(npost_trigger, "Not enough obs after trigger")
                continue

            # make sure the points are actual detections, not saturated or non-detections 
            nd_ind  = (tphotflag == constants.NONDETECT_PHOTFLAG)
            sat_ind = (tphotflag == constants.BAD_PHOTFLAG)
            good_ind = ~(nd_ind | sat_ind)
            ngood = len(ttime[good_ind])
            if ngood <= 3:
                print(ngood, "not enough good observations")
                continue 

            # use only unsaturated measurements
            use_ind = ~sat_ind 
            x  = ttime[use_ind] - trigger_time
            y  = tFluxUnred[use_ind]
            dy = tFluxErrUnred[use_ind]

            # hardcode some limits on parameters
            limit_A = (1E-6,  np.percentile(y, 99))
            limit_B = (-2000., 2000.)
            limit_t0 = (-10., 1.)
            limit_trise = (1., 50.)
            limit_tfall = (1., 50.)

            # define functions to get flux and return likelihood
            def _bazin_model(A, B, t0, trise, tfall):
                """
                Simple parametric model from Bazin et al '09
                https://www.aanda.org/articles/aa/pdf/2009/21/aa11847-09.pdf
                Eqn. 1
                """
                f = A*(np.exp(-(x-t0)/tfall)/(1.+np.exp((x-t0)/trise))) + B
                return f

            def _bazin_likelihood(A, B, t0, trise, tfall):
                f = _bazin_model(A, B, t0, trise, tfall)
                chisq = np.sum(((y - f)**2.)/dy**2.)
                return chisq

            # construct Minuit instance
            m = Minuit(_bazin_likelihood,\
                    A= np.percentile(y, 75), B=0., t0=0., trise=15., tfall=20.,\
                    limit_A=limit_A, limit_B=limit_B, limit_t0=limit_t0, limit_trise=limit_trise, limit_tfall=limit_tfall,\
                    print_level=1)

            m.migrad()

            vals = m.values 
            tparams = []
            for param in m.parameters:
                tparams += [m.values[param], m.errors[param],]
            tparams += [m.fval,]
            bazin_params[pb] = tparams 
        return bazin_params



                





